#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Formulas that encode coloring related problems
"""

from cnfformula.cnf import CNF

from cnfformula.graphs import enumerate_vertices, enumerate_edges, neighbors
from itertools import combinations, combinations_with_replacement, product


def DominatingSet(G, d, alternative=False):
    r"""Generates the clauses for a dominating set for G of size <= d

    The formula encodes the fact that the graph :math:`G` has
    a dominating set of size :math:`d`. This means that it is possible
    to pick at most :math:`d` vertices in :math:`V(G)` so that all remaining
    vertices have distance at most one from the selected ones.

    Parameters
    ----------
    G : networkx.Graph
        a simple undirected graph
    d : a positive int
        the size limit for the dominating set
    alternative : bool
        use an alternative construction that 
        is provably hard from resolution proofs.

    Returns
    -------
    CNF
       the CNF encoding for dominating of size :math:`\leq d` for graph :math:`G`

    """
    # Describe the formula
    description = "{}-dominating set".format(d)

    if hasattr(G, 'name'):
        description += " on " + G.name

    F = CNF(description=description)

    if not isinstance(d, int) or d < 1:
        ValueError("Parameter \"d\" is expected to be a positive integer")

    # Fix the vertex order
    V = enumerate_vertices(G)

    def D(v):
        return "x_{{{0}}}".format(v)

    def M(v, i):
        return "g_{{{0},{1}}}".format(v, i)

    def N(v):
        return tuple(sorted([v] + [u for u in G.neighbors(v)]))

    # Create variables
    for v in V:
        F.add_variable(D(v))
    for i, v in product(range(1, d + 1), V):
        F.add_variable(M(v, i))

    # No two (active) vertices map to the same index
    if alternative:
        for u, v in combinations(V, 2):
            for i in range(1, d + 1):
                F.add_clause([(False, D(u)), (False, D(v)), (False, M(u, i)),
                              (False, M(v, i))])
    else:
        for i in range(1, d + 1):
            for c in CNF.less_or_equal_constraint([M(v, i) for v in V], 1):
                F.add_clause(c)

    # (Active) Vertices in the sequence are not repeated
    if alternative:
        for v in V:
            for i, j in combinations(range(1, d + 1), 2):
                F.add_clause([(False, D(v)), (False, M(v, i)), (False, M(v,
                                                                         j))])
    else:
        for i, j in combinations_with_replacement(range(1, d + 1), 2):
            i, j = min(i, j), max(i, j)
            for u, v in combinations(V, 2):
                u, v = max(u, v), min(u, v)
                F.add_clause([(False, M(u, i)), (False, M(v, j))])

    # D(v) = M(v,1) or M(v,2) or ... or M(v,d)
    if not alternative:
        for i, v in product(range(1, d + 1), V):
            F.add_clause([(False, M(v, i)), (True, D(v))])
    for v in V:
        F.add_clause([(False, D(v))] + [(True, M(v, i))
                                        for i in range(1, d + 1)])

    # Every neighborhood must have a true D variable
    neighborhoods = sorted(set(N(v) for v in V))
    for N in neighborhoods:
        F.add_clause([(True, D(v)) for v in N])

    return F



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of some graph formulas helpers

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

import sys

from cnfformula import PebblingFormula
from cnfformula import StoneFormula
from cnfformula import SparseStoneFormula

from .graph_cmdline import DirectedAcyclicGraphHelper
from .graph_cmdline import BipartiteGraphHelper

from .formula_helpers import FormulaHelper


class PebblingCmdHelper(FormulaHelper):
    """Command line helper for pebbling formulas
    """
    name='peb'
    description='pebbling formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for pebbling formulas

        Arguments:
        - `parser`: parser to load with options.
        """
        DirectedAcyclicGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build the pebbling formula

        Arguments:
        - `args`: command line options
        """
        D= DirectedAcyclicGraphHelper.obtain_graph(args)
        try:
            return PebblingFormula(D)
        except ValueError as e:
            print("\nError: {}".format(e),file=sys.stderr)
            sys.exit(-1)

class StoneCmdHelper(FormulaHelper):
    """Command line helper for stone formulas
    """
    name='stone'
    description='stone formula'
    __doc__ = StoneFormula.__doc__
    
    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for stone formulas

        Arguments:
        - `parser`: parser to load with options.
        """
        DirectedAcyclicGraphHelper.setup_command_line(parser)
        parser.add_argument('s',metavar='<s>',type=int,help="number of stones")

    @staticmethod
    def build_cnf(args):
        """Build the pebbling formula

        Arguments:
        - `args`: command line options
        """
        D= DirectedAcyclicGraphHelper.obtain_graph(args)
        try:
            return StoneFormula(D,args.s)
        except ValueError as e:
            print("\nError: {}".format(e),file=sys.stderr)
            sys.exit(-1)

class SparseStoneCmdHelper(FormulaHelper):
    """Command line helper for stone formulas
    """
    name='stonesparse'
    description='stone formula (sparse version)'
    __doc__ = SparseStoneFormula.__doc__
    
    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for stone formulas

        Arguments:
        - `parser`: parser to load with options.
        """
        DirectedAcyclicGraphHelper.setup_command_line(parser)
        BipartiteGraphHelper.setup_command_line(parser,suffix="_mapping")

    @staticmethod
    def build_cnf(args):
        """Build the pebbling formula

        Arguments:
        - `args`: command line options
        """
        D= DirectedAcyclicGraphHelper.obtain_graph(args)
        B= BipartiteGraphHelper.obtain_graph(args,suffix="_mapping")
        try:
            return SparseStoneFormula(D,B)
        except ValueError as e:
            print("\nError: {}".format(e),file=sys.stderr)
            sys.exit(-1)




#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Init code of the cnfformula pacakge

Essentially it makes visible the names of the formulas and
transformations implemented, plus some IO functions.

"""

# Basic CNF object
from cnfformula.cnf import CNF

# IO functions
from cnfformula.utils.parsedimacs import readCNF
from cnfformula.graphs import readGraph, writeGraph
from cnfformula.graphs import supported_formats as supported_graph_formats

# SAT solvers
from cnfformula.utils.solver import supported_satsolvers
from cnfformula.utils.solver import some_solver_installed

# Formula families implemented
from cnfformula.families.cliquecoloring import CliqueColoring
from cnfformula.families.coloring import GraphColoringFormula
from cnfformula.families.coloring import EvenColoringFormula
from cnfformula.families.counting import CountingPrinciple
from cnfformula.families.counting import PerfectMatchingPrinciple
from cnfformula.families.dominatingset import DominatingSet
from cnfformula.families.graphisomorphism import GraphIsomorphism
from cnfformula.families.graphisomorphism import GraphAutomorphism
from cnfformula.families.ordering import OrderingPrinciple
from cnfformula.families.ordering import GraphOrderingPrinciple
from cnfformula.families.pebbling import PebblingFormula
from cnfformula.families.pebbling import StoneFormula
from cnfformula.families.pebbling import SparseStoneFormula
from cnfformula.families.pigeonhole import PigeonholePrinciple
from cnfformula.families.pigeonhole import GraphPigeonholePrinciple
from cnfformula.families.pigeonhole import BinaryPigeonholePrinciple
from cnfformula.families.ramsey import PythagoreanTriples
from cnfformula.families.ramsey import RamseyLowerBoundFormula
from cnfformula.families.randomformulas import RandomKCNF
from cnfformula.families.subgraph import SubgraphFormula
from cnfformula.families.subgraph import CliqueFormula
from cnfformula.families.subgraph import BinaryCliqueFormula
from cnfformula.families.subgraph import RamseyWitnessFormula
from cnfformula.families.subsetcardinality import SubsetCardinalityFormula
from cnfformula.families.tseitin import TseitinFormula
from cnfformula.families.pitfall import PitfallFormula
from cnfformula.families.cpls import CPLSFormula

# Formula transformation implemented
from cnfformula.transformations.substitutions import AllEqualSubstitution
from cnfformula.transformations.substitutions import ExactlyOneSubstitution
from cnfformula.transformations.substitutions import FlipPolarity
from cnfformula.transformations.substitutions import FormulaLifting
from cnfformula.transformations.substitutions import IfThenElseSubstitution
from cnfformula.transformations.substitutions import MajoritySubstitution
from cnfformula.transformations.substitutions import NotAllEqualSubstitution
from cnfformula.transformations.substitutions import OrSubstitution
from cnfformula.transformations.substitutions import VariableCompression
from cnfformula.transformations.substitutions import XorSubstitution
from cnfformula.transformations.shuffle import Shuffle



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Cnf formulas shuffling."""

import os
import sys
import random
import argparse

from cnfformula import readCNF
from cnfformula import Shuffle

from .cmdline import paginate_or_redirect_stdout
from .cmdline import redirect_stdin
from .cmdline import setup_SIGINT
from .cmdline import CLIParser, CLIError

from .msg import interactive_msg
from .msg import error_msg
from .msg import msg_prefix
from .msg import InternalBug


def cli(argv=sys.argv, mode='output'):
    """CNFgen shuffler 

    This function provide the main interface to cnfshuffle.


    Parameters
    ----------
    argv: list, optional
        The list of token with the command line arguments/options.

    mode: str
        One among 'formula', 'string', 'output' (latter is the default)
        - 'formula' return a CNF object
        - 'string' return the string with the output of CNFgen
        - 'output' output the formula to the user
    """

    # Parse the command line arguments
    progname = os.path.basename(argv[0])
    parser = CLIParser(prog=progname,
                       description="""
    Reshuffle the input CNF. Returns a formula logically
    equivalent to the input with random application of
    (1) Polarity flips (2) Variables permutation (3) Clauses permutation.
    """,
                       epilog="""
    For more information type '%s [--help | -h ]'
    """ % (progname))

    parser.add_argument('--output',
                        '-o',
                        type=argparse.FileType('w'),
                        metavar="<output>",
                        default='-',
                        help="""Output file. The formula is saved
                        on file instead of being sent to standard
                        output. Setting '<output>' to '-' is another
                        way to send the formula to standard output.
                        (default: -)
                        """)
    parser.add_argument('--seed',
                        '-S',
                        metavar="<seed>",
                        default=None,
                        type=str,
                        action='store',
                        help="""Seed for any random process in the
                        program. Any python hashable object will
                        be fine.  (default: current time)
                        """)
    parser.add_argument(
        '--input',
        '-i',
        type=argparse.FileType('r'),
        metavar="<input>",
        default='-',
        help=
        """Input file. A formula in dimacs format. Setting '<input>' to '-' is
                        another way to read from standard input.
                        (default: -)
                        """)
    parser.add_argument('--no-polarity-flips',
                        '-p',
                        action='store_true',
                        dest='no_polarity_flips',
                        help="No polarity flips")
    parser.add_argument('--no-variables-permutation',
                        '-v',
                        action='store_true',
                        dest='no_variable_permutations',
                        help="No permutation of variables")
    parser.add_argument('--no-clauses-permutation',
                        '-c',
                        action='store_true',
                        dest='no_clause_permutations',
                        help="No permutation of clauses")
    parser.add_argument('--quiet',
                        '-q',
                        action='store_false',
                        default=True,
                        dest='verbose',
                        help="""Output just the formula with no header.""")

    # Be lenient on non string arguments
    argv = [str(x) for x in argv]

    # Process the options
    args = parser.parse_args(argv[1:])

    # If necessary, init the random generator
    if hasattr(args, 'seed') and args.seed:
        random.seed(args.seed)

    msg = """Waiting for a DIMACS formula on <stdin>.
             Alternatively you can feed a formula to <stdin>
             with piping or using '-i' command line argument."""

    with redirect_stdin(args.input), msg_prefix('c '):
        with msg_prefix("INPUT: "):
            interactive_msg(msg, filltext=70)

        F = readCNF()

    # Default permutation
    if not args.no_variable_permutations:
        variable_permutation = None
    else:
        variable_permutation = list(F.variables())

    if not args.no_clause_permutations:
        clause_permutation = None
    else:
        clause_permutation = list(range(len(F)))

    if not args.no_polarity_flips:
        polarity_flip = None
    else:
        polarity_flip = [1] * len(list(F.variables()))

    G = Shuffle(F, variable_permutation, clause_permutation, polarity_flip)

    if mode == 'formula':
        return G
    elif mode == 'string':
        return G.dimacs(export_header=args.verbose)
    else:
        with paginate_or_redirect_stdout(args.output):
            G._dimacs_dump_clauses(output=sys.stdout,
                                   export_header=args.verbose)


# Launcher
def main():
    setup_SIGINT()

    try:

        cli(sys.argv)

    except ValueError as e:
        error_msg("DIMACS ERROR: " + str(e))
        sys.exit(-1)

    except CLIError as e:
        error_msg(str(e))
        sys.exit(-1)

    except InternalBug as e:
        print(str(e), file=sys.stderr)
        sys.exit(-1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of the pigeonhole principle formulas
"""

from cnfformula.cnf import CNF
from cnfformula.graphs import is_dag, enumerate_vertices
from cnfformula.graphs import has_bipartition, bipartite_sets

from itertools import product
from collections import OrderedDict

import sys


def _uniqify_list(seq):
    """Remove duplicates while maintaining the order.

    (due to Dave Kirby) 

    Seen on https://www.peterbe.com/plog/uniqifiers-benchmark
    """
    seen = set()
    return [x for x in seq if x not in seen and not seen.add(x)]


def PebblingFormula(digraph):
    """Pebbling formula

    Build a pebbling formula from the directed graph. If the graph has
    an `ordered_vertices` attribute, then it is used to enumerate the
    vertices (and the corresponding variables).

    Arguments:
    - `digraph`: directed acyclic graph.
    """
    if not is_dag(digraph):
        raise ValueError(
            "Pebbling formula is defined only for directed acyclic graphs")

    description = 'Pebbling formula'
    if hasattr(digraph, 'name'):
        description += " for " + digraph.name

    peb = CNF(description=description)

    # add variables in the appropriate order
    vertices = enumerate_vertices(digraph)
    position = dict((v, i) for (i, v) in enumerate(vertices))

    for v in vertices:
        peb.add_variable(
            v, description="There is a pebble on vertex ${}$".format(v))

    # add the clauses
    for v in vertices:

        # If predecessors are pebbled the vertex must be pebbled
        pred = sorted(digraph.predecessors(v), key=lambda x: position[x])
        peb.add_clause_unsafe([(False, p) for p in pred] + [(True, v)])

        if digraph.out_degree(v) == 0:  #the sink
            peb.add_clause_unsafe([(False, v)])

    return peb


def stone_formula_helper(F, D, mapping):
    """Stones formulas helper

    Builds the clauses of a stone formula given the mapping object
    between stones and vertices of the DAG. THis is not supposed to be
    called by user code, and indeed this function assumes the following facts.

    - :math:`D` is equal to `mapping.domain()`
    - :math:`D` is a DAG

    Parameters
    ----------
    F : CNF
        the CNF which will contain the clauses of the stone formula

    D : a directed acyclic graph
        it should be a directed acyclic graph.

    mapping : unary_mapping object
       mapping between stones and graph vertices


    See Also
    --------
    StoneFormula : the classic stone formula
    SparseStoneFormula : stone formula with sparse mapping
    """

    # add variables in the appropriate order
    vertices = enumerate_vertices(D)

    # caching variable names
    color_vn = {}
    stone_vn = {}

    # Stones->Vertices variables
    for v in mapping.variables():
        F.add_variable(v)
    for v in vertices:
        for j in mapping.images(v):
            stone_vn[(v, j)] = mapping.var_name(v, j)

    # Color variables
    for stone in mapping.range():
        color_vn[stone] = "R_{{{0}}}".format(stone)
        F.add_variable(color_vn[stone],
                       description="Stone ${}$ is red".format(stone))

    # Each vertex has some stone
    for cls in mapping.clauses():
        F.add_clause_unsafe(cls)

    # If predecessors have red stones, the sink must have a red stone
    for v in vertices:
        for j in mapping.images(v):
            pred = sorted(D.predecessors(v),
                          key=lambda x: mapping.RankDomain[x])
            for stones_tuple in product(*tuple(
                [s for s in mapping.images(v) if s != j] for v in pred)):
                F.add_clause_unsafe([(False, stone_vn[(p, s)])
                                     for (p, s) in zip(pred, stones_tuple)] +
                                    [(False, stone_vn[(v, j)])] +
                                    [(False, color_vn[s])
                                     for s in _uniqify_list(stones_tuple)] +
                                    [(True, color_vn[j])])

        if D.out_degree(v) == 0:  #the sink
            for j in mapping.images(v):
                F.add_clause_unsafe([(False, stone_vn[(v, j)]),
                                     (False, color_vn[j])])

    return F


def StoneFormula(D, nstones):
    """Stone formulas

    The stone formulas have been introduced in [2]_ and generalized in
    [1]_. They are one of the classic examples that separate regular
    resolutions from general resolution [1]_.

    A \"Stones formula\" from a directed acyclic graph :math:`D`
    claims that each vertex of the graph is associated with one on
    :math:`s` stones (not necessarily in an injective way).
    In particular for each vertex :math:`v` in :math:`V(D)` and each
    stone :math:`j` we have a variable :math:`P_{v,j}` that claims
    that stone :math:`j` is associated to vertex :math:`v`.

    Each stone can be either red or blue, and not both.
    The propositional variable :math:`R_j` if true when the stone
    :math:`j` is red and false otherwise.

    The clauses of the formula encode the following constraints.
    If a stone is on a source vertex (i.e. a vertex with no incoming
    edges), then it must be red. If all stones on the predecessors of
    a vertex are red, then the stone of the vertex itself must be red.

    The formula furthermore enforces that the stones on the sinks
    (i.e. vertices with no outgoing edges) are blue.

    .. note:: The exact formula structure depends by the graph and on
              its topological order, which is determined by the
              ``enumerate_vertices(D)``.

    Parameters
    ----------
    D : a directed acyclic graph
        it should be a directed acyclic graph.
    nstones : int
       the number of stones.

    Raises
    ------
    ValueError
       if :math:`D` is not a directed acyclic graph
    
    ValueError
       if the number of stones is negative

    References
    ----------
    .. [1] M. Alekhnovich, J. Johannsen, T. Pitassi and A. Urquhart
    	   An Exponential Separation between Regular and General Resolution.
           Theory of Computing (2007)
    .. [2] R. Raz and P. McKenzie
           Separation of the monotone NC hierarchy.
           Combinatorica (1999)

    """
    if not is_dag(D):
        raise ValueError(
            "Stone formulas are defined only for directed acyclic graphs.")

    if nstones < 0:
        raise ValueError("There must be at least one stone.")

    if hasattr(D, 'name'):
        description = "Stone formula of {} with {} stones".format(
            D.name, nstones)
    else:
        description = "Stone formula with {} stones".format(nstones)

    cnf = CNF(description=description)

    # Add variables in the appropriate order
    vertices = enumerate_vertices(D)
    position = dict((v, i) for (i, v) in enumerate(vertices))
    stones = list(range(1, nstones + 1))

    # Caching variable names
    color_vn = {}
    stone_vn = {}

    # Stones->Vertices variables
    for v in vertices:
        for j in stones:
            stone_vn[(v, j)] = "P_{{{0},{1}}}".format(v, j)
            cnf.add_variable(stone_vn[(v, j)],
                             description="Stone ${1}$ on vertex ${0}$".format(
                                 v, j))

    # Color variables
    for j in stones:
        color_vn[j] = "R_{{{0}}}".format(j)
        cnf.add_variable(color_vn[j],
                         description="Stone ${}$ is red".format(j))

    # Each vertex has some stone
    for v in vertices:
        cnf.add_clause_unsafe([(True, stone_vn[(v, j)]) for j in stones])

    # If predecessors have red stones, the sink must have a red stone
    for v in vertices:
        for j in stones:
            pred = sorted(D.predecessors(v), key=lambda x: position[x])
            for stones_tuple in product([s for s in stones if s != j],
                                        repeat=len(pred)):
                cnf.add_clause_unsafe([(False, stone_vn[(p, s)])
                                       for (p, s) in zip(pred, stones_tuple)] +
                                      [(False, stone_vn[(v, j)])] +
                                      [(False, color_vn[s])
                                       for s in _uniqify_list(stones_tuple)] +
                                      [(True, color_vn[j])])

        if D.out_degree(v) == 0:  #the sink
            for j in stones:
                cnf.add_clause_unsafe([(False, stone_vn[(v, j)]),
                                       (False, color_vn[j])])

    return cnf


def SparseStoneFormula(D, B):
    """Sparse Stone formulas

    This is a variant of the :py:func:`StoneFormula`. See that for
    a description of the formula. This variant is such that each
    vertex has only a small selection of which stone can go to that
    vertex. In particular which stones are allowed on each vertex is
    specified by a bipartite graph :math:`B` on which the left
    vertices represent the vertices of DAG :math:`D` and the right
    vertices are the stones. 

    If a vertex of :math:`D` correspond to the left vertex :math:`v`
    in :math:`B`, then its neighbors describe which stones are allowed
    for it.

    The vertices in :math:`D` do not need to have the same name as the
    one on the left side of :math:`B`. It is only important that the
    number of vertices in :math:`D` is the same as the vertices in the
    left side of :math:`B`.

    In that case the element at position :math:`i` in the ordered
    sequence ``enumerate_vertices(D)`` corresponds to the element of
    rank :math:`i` in the sequence of left side vertices of
    :math:`B` according to the output of ``Left, Right =
    bipartite_sets(B)``.

    Standard :py:func:`StoneFormula` is essentially equivalent to
    a sparse stone formula where :math:`B` is the complete graph.

    Parameters
    ----------
    D : a directed acyclic graph
        it should be a directed acyclic graph.
    B : bipartite graph

    Raises
    ------
    ValueError
       if :math:`D` is not a directed acyclic graph
    
    ValueError
       if :math:`B` is not a bipartite graph

    ValueError
       when size differs between :math:`D` and the left side of
       :math:`B`

    See Also
    --------
    StoneFormula

    """
    if not is_dag(D):
        raise ValueError(
            "Stone formulas are defined only for directed acyclic graphs.")

    if not has_bipartition(B):
        raise ValueError(
            "Vertices to stones mapping must be specified with a bipartite graph"
        )

    Left, Right = bipartite_sets(B)
    nstones = len(Right)

    if len(Left) != D.order():
        raise ValueError(
            "Formula requires the bipartite left side to match #vertices of the DAG."
        )

    if hasattr(D, 'name'):
        description = "Sparse stone formula of {} with {} stones".format(
            D.name, nstones)
    else:
        description = "Sparse stone formula with {} stones".format(nstones)

    cnf = CNF(description=description)

    # add variables in the appropriate order
    vertices = enumerate_vertices(D)
    stones = list(range(1, nstones + 1))

    mapping = cnf.unary_mapping(vertices, stones, sparsity_pattern=B)

    stone_formula_helper(cnf, D, mapping)
    return cnf



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""Utilities to build CNF formulas interesting for proof complexity.

The module `cnfgen`  contains facilities to generate  cnf formulas, in
order to  be printed  in dimacs  or LaTeX  formats. Such  formulas are
ready to be  fed to sat solvers.  In particular  the module implements
both a library of CNF generators and a command line utility.

Create the CNFs:

>>> from cnfformula import CNF
>>> c=CNF([ [(True,"x1"),(True,"x2"),(False,"x3")], \
          [(False,"x2"),(True,"x4")] ])
>>> print( c.dimacs(export_header=False) )
p cnf 4 2
1 2 -3 0
-2 4 0

You can add clauses later in the process:

>>> c.add_clause( [(False,"x3"),(True,"x4"),(False,"x5")] )
>>> print( c.dimacs(export_header=False))
p cnf 5 3
1 2 -3 0
-2 4 0
-3 4 -5 0

"""

import os
import sys
import random
import io
import argparse

from cnfformula.info import info

from .cmdline import paginate_or_redirect_stdout
from .cmdline import setup_SIGINT
from .cmdline import CLIParser, CLIError

from .cmdline import get_formula_helpers
from .cmdline import get_transformation_helpers

from .msg import error_msg
from .msg import msg_prefix
from .msg import InternalBug

#################################################################
#          Command line tool follows
#################################################################

# Help strings
usage_string = """{} [-h] [-V] [--output <output>]
                 [--output-format {{latex,dimacs}}] [--seed <seed>]
                 [--verbose | --quiet] [--tutorial]
                 <formula> <args> ...
                 [-T <transformation> <args>]
                 [-T <transformation> <args>]
                 ..."""

description_string = """example:
 {0} php 100 40         --- Pigeonhole principle 100 pigeons 40 holes (unsat)
 {0} op  14             --- Ordering principle on 14 elements (unsat)
 {0} randkcnf 3 10 5    --- Random 3-CNF with 10 vars and 5 clauses

tutorial:
 {0} --tutorial         --- shows a command line tutorial

"""

tutorial_string = """
                 CNFGEN TUTORIAL

{0} builds CNF formulas mostly coming from proof complexity
literature, to use as benchmark against SAT solvers. Basic usage is

    {0} <formula> <arg1> <arg2> ...

which builds a CNF from family <formula> with various parameters.
For example a Pigeonhole principle formula from 5 to 4 pigeons can be
build with command line

    {0} php 5 4

Various transformations can be applied to the generated formula, one
after the other. For example.

    {0} php 5 4 -T shuffle -T xor 3

Create a pigeonhole principle formula first, then applies the
'shuffle' transformation and finally the 'xor' transformation with
parameter 3.

Some formulas are built on top of graph structures, passed as input
files or randomly generated. The command lines

    {0} tseitin -i graph.dot
    {0} tseitin --gnd 10 4

build Tseitin formulas, respectively, over a graph passed as a DOT
file, and over a random 4-regular graph of 10 vertices.

For the full list of formulas and formula transformations type one of

    {0} -h
    {0} --help

For the help of a specific CNF family named <formula> type one of

    {0} <formula> -h
    {0} <formula> --help

For the help of a specific formula transformation pick any CNF family
<formula> and type one of

    {0} <formula> <args> -T <transformation> -h
    {0} <formula> <args> -T <transformation> --help

where <transformation> is the name of the formula transformation to
get the help for.
"""


def setup_command_line_parsers(progname, fhelpers, thelpers):
    """Create the parser for formula and transformation arguments.

    General options
    - query version
    - verbosity/silence
    - outputfile
    - outputformat
    - random seed

    Formula options via subcommands
    - one for each formula helper

    Parameters:
    -----------
    progname:
        the name of the program
    fhelpers:
        the cmdline helpers for the formulas
    thelpers:
        the cmdline helpers for the transformations

    Return:
    -------
    two parsers, one for formula commands and one for transformation commands.
    """

    # First we setup the parser for transformation command lines
    t_parser = CLIParser(add_help=False,
                         formatter_class=argparse.RawDescriptionHelpFormatter)

    t_subparsers = t_parser.add_subparsers(
        prog=progname + " <formula> <args> -T",
        title="Available formula transformation",
        metavar="<transformation>")
    for sc in thelpers:
        p = t_subparsers.add_parser(
            sc.name,
            help=sc.description,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        sc.setup_command_line(p)
        sc.subparser = p
        p.set_defaults(transformation=sc)

    # now we setup the main parser for the formula generation command
    parser = CLIParser(prog=progname,
                       usage=usage_string.format(progname),
                       description=description_string.format(progname),
                       formatter_class=argparse.RawDescriptionHelpFormatter)

    class PrintTutorial(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            with paginate_or_redirect_stdout(sys.stdout):
                print(tutorial_string.format(progname))
            sys.exit(os.EX_OK)

    parser.add_argument('-V',
                        '--version',
                        action='version',
                        version='{project} ({version})'.format(**info))
    parser.add_argument('--tutorial',
                        nargs=0,
                        action=PrintTutorial,
                        help="show a brief tutorial")
    parser.add_argument(
        '--output',
        '-o',
        type=argparse.FileType('w', encoding='utf-8'),
        metavar="<output>",
        default='-',
        help="""Save the formula to <output>. Setting '<output>' to '-' sends the
                        formula to standard output. (default: -)
                        """)
    parser.add_argument('--output-format',
                        '-of',
                        choices=['latex', 'dimacs'],
                        default='dimacs',
                        help="""
                        Output format of the formulas. 'latex' is
                        convenient to insert formulas into papers, and
                        'dimacs' is the format used by sat solvers.
                        (default: dimacs)
                        """)

    parser.add_argument('--seed',
                        '-S',
                        metavar="<seed>",
                        default=None,
                        type=str,
                        action='store',
                        help="""Seed for any random process in the
                        program. (default: current time)
                        """)
    g = parser.add_mutually_exclusive_group()
    g.add_argument('--verbose',
                   '-v',
                   action='store_true',
                   default=True,
                   help="""Output formula header and comments.""")
    g.add_argument('--quiet',
                   '-q',
                   action='store_false',
                   dest='verbose',
                   help="""Output just the formula with no header.""")

    # setup each formula command parser
    subparsers = parser.add_subparsers(prog=progname,
                                       title="Available formula types",
                                       metavar='<formula>')
    for sc in fhelpers:
        p = subparsers.add_parser(
            sc.name,
            help=sc.description,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        sc.setup_command_line(p)
        sc.subparser = p
        p.set_defaults(generator=sc)

    # Attach the list of available transformations
    # but without usage string
    extension = t_parser.format_help().splitlines()
    extension = "\n".join(extension[2:])
    parser.epilog = extension

    return parser, t_parser


def parse_command_line(argv, fparser, tparser):
    """Parser command line arguments

    Split command line around '-T' options in parts. Extract the
    formula generator setup from the first part, and extract a chain
    of formula transformation commands

    Parameters:
    -----------
    argv:
        command line arguments
    parser:
        parser for the formulas commands
    t_parser:
        parser for the transformations commands

    Return:
    -------
    fargs, list(targs)
        setup for the various commands
    """
    # Split the command line
    cmd_chunks = [[]]
    for arg in argv:
        if arg == '-T':
            cmd_chunks.append([])
        else:
            cmd_chunks[-1].append(arg)

    generator_cmd = cmd_chunks[0][1:]
    transformation_cmds = cmd_chunks[1:]

    # Parse the various component of the command line
    fargs = fparser.parse_args(generator_cmd)
    targs = []
    for cmd in transformation_cmds:
        targs.append(tparser.parse_args(cmd))

    return fargs, targs


def build_latex_cmdline_description(argv, args, t_args):
    """Build the latex documentation of the components of the formula.

    Insert additonal latex documentation of the formula. For example
    it includes the input files used to build it.

    Parameters
    ----------
    argv : list(str)
        arguments on command line
    args:
        parsed arguments for the formula
    t_args: list
        group of parsed arguments for the transformations

    """
    # The docstring of the formula family generator
    cmdline_descr = []
    if hasattr(args.generator, "docstring"):
        cmdline_descr += [
            "\\noindent\\textbf{Docstring:}",
            "\\begin{lstlisting}[breaklines,basicstyle=\\small]",
            args.generator.docstring, "\\end{lstlisting}"
        ]

    # Find input files specified in the command line
    input_files = []
    for namespace in [args] + t_args:
        data = iter(vars(namespace).items())
        data = [v for k, v in data if not k.startswith("_")]
        data = [f for f in data if isinstance(f, io.IOBase)]
        data = [f for f in data if (f != sys.stdin) and f.mode == 'r']
        input_files.extend(data)

    # Add them all to the latex document
    for f in input_files:
        f.seek(0, 0)
        cmdline_descr += [
            "\\noindent\\textbf{Input file} \\verb|%s|" % f.name,
            "\\begin{lstlisting}[breaklines,basicstyle=\\small]",
            f.read(), "\\end{lstlisting}"
        ]

    # Return all as a single string
    cmdline_descr += ['\n']
    return "\n".join(cmdline_descr)


# Command line interface
def cli(argv=sys.argv, mode='output'):
    """CNFgen main command line interface

    This function provide the main interface to CNFgen. It sets up the
    command line, parses the command line arguments, builds the
    appropriate formula and outputs its representation.


    Parameters
    ----------
    argv: list, optional
        The list of token with the command line arguments/options.

    mode: str
        One among 'formula', 'string', 'output' (latter is the default)
        - 'formula' return a CNF object
        - 'string' return the string with the output of CNFgen
        - 'output' output the formula to the user

    Raises
    ------
    CLIError:
        The command line arguments are wrong or inconsistent

    Return
    ------
    depends on the 'mode' argument
    """

    progname = os.path.basename(argv[0])

    formula_helpers = get_formula_helpers()
    transformation_helpers = get_transformation_helpers()

    parser, t_parser = setup_command_line_parsers(progname, formula_helpers,
                                                  transformation_helpers)

    # Be lenient on non string arguments
    argv = [str(x) for x in argv]
    with msg_prefix('c '):
        args, t_args = parse_command_line(argv, parser, t_parser)

    # Correctly infer the comment character, useful to shield
    # the output.
    comment_char = {'dimacs': 'c ', 'latex': '% '}
    try:
        cprefix = comment_char[args.output_format]
    except KeyError:
        raise InternalBug("Unknown output format")

    with msg_prefix(cprefix):

        # Check basics: formula family and transformation specs
        if not hasattr(args, 'generator'):
            parser.error(
                "You did not tell which formula you wanted to generate.\n")

        for argdict in t_args:
            if not hasattr(argdict, 'transformation'):
                t_parser.error(
                    "You used option '-T' but did not pick a transformation.\n"
                )

        # Generate the formula and apply transformations
        if hasattr(args, 'seed') and args.seed:
            random.seed(args.seed)

        try:
            cnf = args.generator.build_cnf(args)
        except (CLIError, ValueError) as e:
            args.generator.subparser.error(e)
        except RuntimeError as e:
            raise InternalBug(e)

        for argdict in t_args:
            try:
                cnf = argdict.transformation.transform_cnf(cnf, argdict)
            except (CLIError, ValueError) as e:
                argdict.transformation.subparser.error(e)
            except RuntimeError as e:
                raise InternalBug(e)

        cnf.header['command line'] = "cnfgen " + " ".join(argv[1:])

        if mode == 'formula':
            return cnf

        # Output
        if args.output_format == 'latex':

            extra_text = build_latex_cmdline_description(argv, args, t_args)

            output = cnf.latex(export_header=args.verbose,
                               extra_text=extra_text,
                               full_document=True)

        elif args.output_format == 'dimacs':

            output = cnf.dimacs(export_header=args.verbose)

        else:
            raise InternalBug("Unknown output format")

    if mode == 'string':
        return output

    with paginate_or_redirect_stdout(args.output):
        print(output)


# command line entry point
def main():

    setup_SIGINT()

    try:

        cli(sys.argv, mode='output')

    except CLIError as e:
        error_msg(e)
        sys.exit(-1)

    except InternalBug as e:
        print(e, file=sys.stderr)
        sys.exit(-1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Utilities to manage graph formats and graph files in order to build
formulas that are graph based.
"""

import copy

__all__ = [
    "supported_formats", "readGraph", "writeGraph", "is_dag",
    "has_bipartition", "bipartite_sets", "enumerate_vertices",
    "enumerate_edges", "neighbors", "bipartite_random_left_regular",
    "bipartite_random_regular", "dag_complete_binary_tree", "dag_pyramid"
]

#################################################################
#          Graph Decoders (first is default)
#################################################################

_graphformats = {
    'dag': ['kthlist', 'gml', 'dot'],
    'digraph': ['kthlist', 'gml', 'dot', 'dimacs'],
    'simple': ['kthlist', 'gml', 'dot', 'dimacs'],
    'bipartite': ['kthlist', 'matrix', 'gml', 'dot']
}


def supported_formats():
    """The graph file formats supported by the library."""
    return copy.deepcopy(_graphformats)


#################################################################
#          Import third party code
#################################################################

import sys
from io import StringIO, BytesIO
import io
import os

import networkx
import networkx.algorithms


def has_dot_library():
    """Test the presence of pydot
    """
    try:
        # newer version of networkx
        from networkx import nx_pydot
        import pydot
        del pydot
        return True
    except ImportError:
        pass

    return False


# Check that DOT is a supported format
if not has_dot_library():
    for k in list(_graphformats.values()):
        try:
            k.remove('dot')
        except ValueError:
            pass

#################################################################
#          Graph reader/writer
#################################################################


def _process_graph_io_arguments(iofile, graph_type, file_format, multi_edges):
    """Test if the argument for the graph I/O functions make sense"""

    # Check the file
    if not isinstance(iofile, io.TextIOBase) and \
       not isinstance(iofile, io.IOBase) and \
       not isinstance(iofile, StringIO):
        raise ValueError(
            "The IO stream \"{}\" does not correspond to a file".format(
                iofile))

    # Check the graph type specification
    if graph_type not in _graphformats:
        raise ValueError("The graph type must be one of " +
                         list(_graphformats.keys()))

    elif graph_type in {"dag", "digraph"}:
        if multi_edges:
            grtype = networkx.MultiDiGraph
        else:
            grtype = networkx.DiGraph
    elif graph_type in {"simple", 'bipartite'}:
        if multi_edges:
            grtype = networkx.MultiGraph
        else:
            grtype = networkx.Graph
    else:
        raise RuntimeError(
            "Unchecked graph type argument: {}".format(graph_type))

    # Check/discover file format specification
    if file_format == 'autodetect':
        try:
            extension = os.path.splitext(iofile.name)[-1][1:]
        except AttributeError:
            raise ValueError(
                "Cannot guess a file format from an IO stream with no name. Please specify the format manually."
            )

        if extension not in _graphformats[graph_type]:
            raise ValueError("Cannot guess a file format for {} graphs from the extension of \"{}\". Please specify the format manually.".\
                             format(graph_type,iofile.name))
        else:
            file_format = extension

    elif file_format not in _graphformats[graph_type]:
        raise ValueError("For {} graphs we only support these formats: ".
                         format(graph_type) + _graphformats[graph_type])

    return (grtype, file_format)


def readGraph(input_file,
              graph_type,
              file_format='autodetect',
              multi_edges=False):
    """Read a Graph from file

    The graph are managed using the NetworkX library, and most of the
    input and output formats are the ones supported by it. Plus we
    added support for some more *hackish* formats that are useful
    in applications.

    for the "simple" and "bipartite" types, the graph is actually
    a (Multi)Graph object, while it is a (Multi)DiGraph in case of
    "dag" or "digraph".

    In the case of "dag" type, the graph is guaranteed to be acyclic
    and to pass the ``is_dag`` test. In the case of "bipartite" type,
    the graph is guaranteed to have the two parts labeled and to pass
    the ``has_bipartition`` test.


    Parameters
    -----------
    input_file: str or file-like object
        the input file from which the graph is read. If it is a string
        then the graph is read from a file with that string as
        filename. Otherwise if the input_file is a file object (or
        a text stream), the graph is read from there.

        Input files are assumed to be UTF-8 by default.

    graph_type: string in {"simple","digraph","dag","bipartite"}
        see also :py:func:`cnfformula.graph.supported_formats`

    file_format: string, optional
        The file format that the parser should expect to receive.
        See also :py:func:`cnfformula.graph.supported_formats`. By default
        it tries to autodetect it from the file name extension (when applicable).

    multi_edges: bool,optional
        are multiple edge allowed in the graph? By default this is not allowed.

    Returns
    -------
    a graph object
        one among networkx.DiGraph, networkx.MultiDiGraph,
        networkx.Graph, networkx.MultiGraph object.

    Raises
    ------
    ValueError
        raised when either ``input_file`` is neither a file object
        nor a string, or when ``graph_type`` and ``file_format`` are
        invalid choices.

    IOError
        it is impossible to read the ``input_file``

    See Also
    --------
    writeGraph, is_dag, has_bipartition

    """

    # file name instead of file object
    if isinstance(input_file, str):
        with open(input_file, 'r', encoding='utf-8') as file_handle:
            return readGraph(file_handle, graph_type, file_format, multi_edges)

    grtype, file_format = _process_graph_io_arguments(input_file, graph_type,
                                                      file_format, multi_edges)

    if file_format == 'dot':

        # This is a workaround. In theory a broken dot file should
        # cause a pyparsing.ParseError but the dot_reader used by
        # networkx seems to mismanage that and to cause a TypeError
        #
        try:
            G = grtype(networkx.nx_pydot.read_dot(input_file))
        except TypeError:
            raise ValueError('Parse Error in dot file')

    elif file_format == 'gml':

        # Networkx's GML reader expects to read from ascii encoded
        # binary file. We could have sent the data to a temporary
        # binary buffer but for some reasons networkx's GML reader
        # function is poorly written and does not like such buffers.
        # It turns out we can pass the data as a list of
        # encoded ascii lines.

        try:
            G = grtype(
                networkx.read_gml(line.encode('ascii') for line in input_file))
        except networkx.NetworkXError as errmsg:
            raise ValueError("[Parse error in GML input] {} ".format(errmsg))
        except UnicodeEncodeError as errmsg:
            raise ValueError(
                "[Non-ascii chars in GML file] {} ".format(errmsg))

    elif file_format == 'kthlist':

        G = _read_graph_kthlist_format(input_file,
                                       grtype,
                                       bipartition=(graph_type == 'bipartite'))

    elif file_format == 'dimacs':

        G = _read_graph_dimacs_format(input_file, grtype)

    elif file_format == 'matrix':

        G = _read_graph_matrix_format(input_file)

    else:
        raise RuntimeError(
            "[Internal error] Format {} not implemented".format(file_format))

    if graph_type == "dag" and not is_dag(G):
        raise ValueError("[Input error] Graph must be acyclic")

    if graph_type == "bipartite" and not has_bipartition(G):
        raise ValueError(
            "[Input error] Graph vertices miss the 'bipartite' 0,1 label.")

    return G


def writeGraph(G, output_file, graph_type, file_format='autodetect'):
    """Write a graph to a file

    Parameters
    -----------
    G : networkx.Graph (or similar)

    output_file: file object
        the output file to which the graph is written. If it is a string
        then the graph is written to a file with that string as
        filename. Otherwise if ``output_file`` is a file object (or
        a text stream), the graph is written there.

        The file is written in UTF-8 by default.

    graph_type: string in {"simple","digraph","dag","bipartite"}
        see also :py:func:`cnfformula.graph.supported_formats`

    file_format: string, optional
        The file format that the parser should expect to receive.
        See also :py:func:`cnfformula.graph.supported_formats`. By default
        it tries to autodetect it from the file name extension (when applicable).

    Returns
    -------
    None

    Raises
    ------
    ValueError
        raised when either ``output_file`` is neither a file object
        nor a string, or when ``graph_type`` and ``file_format`` are
        invalid choices.

    IOError
        it is impossible to write on the ``output_file``

    See Also
    --------
    readGraph

    """

    # file name instead of file object
    if isinstance(output_file, str):
        with open(output_file, 'w', encoding='utf-8') as file_handle:
            return writeGraph(G, file_handle, graph_type, file_format)

    _, file_format = _process_graph_io_arguments(output_file, graph_type,
                                                 file_format, False)

    if file_format == 'dot':

        networkx.nx_pydot.write_dot(G, output_file)

    elif file_format == 'gml':

        # Networkx's GML writer expects to write to an ascii encoded
        # binary file. Thus we need to let Networkx write to
        # a temporary binary ascii encoded buffer and then convert the
        # content before sending it to the output file.
        tempbuffer = io.BytesIO()
        networkx.write_gml(G, tempbuffer)
        print(tempbuffer.getvalue().decode('ascii'), file=output_file)

    elif file_format == 'kthlist':

        _write_graph_kthlist_format(G,
                                    output_file,
                                    bipartition=(graph_type == 'bipartite'))

    elif file_format == 'dimacs':

        _write_graph_dimacs_format(G, output_file)

    elif file_format == 'matrix':

        _write_graph_matrix_format(G, output_file)

    else:
        raise RuntimeError(
            "[Internal error] Format {} not implemented".format(file_format))


#
# test for dag / with caching
#
def is_dag(digraph):
    """Test is a directed graph is acyclic

    if the input graph has a member `topologically_sorted' then assumed that
    there is a member `ordered_vertices' and that it is a topological order.

    Arguments:
    - `digraph`: input graph
    """

    if not isinstance(digraph, (networkx.MultiDiGraph, networkx.DiGraph)):
        return False

    elif hasattr(digraph, "topologically_sorted"):

        assert hasattr(digraph, "ordered_vertices")
        assert digraph.order() == len(digraph.ordered_vertices)
        assert set(digraph.nodes()) == set(digraph.ordered_vertices)
        assert networkx.algorithms.is_directed_acyclic_graph(digraph)
        return True

    else:
        return networkx.algorithms.is_directed_acyclic_graph(digraph)


def has_bipartition(G):
    """Check that the graph is labelled with a bipartition

    NetworkX follows the convention that bipartite graphs have their
    vertices labeled with the bipartition. In particular each vertex
    has the 'bipartite' attribute with is either 0 or 1.

    Parameters
    ----------
    G: networkx.Graph

    """
    try:
        for n in G.nodes():
            # Accept both string and int representation
            if not G.nodes[n]['bipartite'] in [0, 1, '0', '1']:
                return False
    except KeyError:
        return False

    return True


def bipartite_sets(G):
    """Return the bipartition of the vertices

    Raises 'ValueError' if bipartition labels are missing"""
    try:
        Left = sorted(
            [v for v, d in G.nodes(data=True) if d['bipartite'] in [0, '0']])
        Right = sorted(
            [v for v, d in G.nodes(data=True) if d['bipartite'] in [1, '1']])
    except KeyError:
        raise ValueError("The encoded bipartition is missing")

    return Left, Right


#
# Use, when possible, a fixed vertex order
#
def enumerate_vertices(graph):
    """Return the ordered list of vertices of `graph`

    Parameters
    ----------
    graph : input graph
    """
    if hasattr(graph, "ordered_vertices"):
        assert graph.order() == len(graph.ordered_vertices)
        assert set(graph.nodes()) == set(graph.ordered_vertices)
        return graph.ordered_vertices
    else:
        setattr(graph, "ordered_vertices", sorted(graph.nodes()))
        return graph.ordered_vertices


def enumerate_edges(graph):
    """Return the ordered list of edges of `graph`

    Parameters
    ----------
    graph : input graph
    """
    if hasattr(graph, "ordered_edges"):
        assert set(graph.edges()) == set(graph.ordered_edges)
        return graph.ordered_edges
    else:
        setattr(graph, "ordered_edges", sorted(graph.edges()))
        return graph.ordered_edges


def neighbors(graph, v):
    """Return the ordered list of neighbors ov a vertex

    Parameters
    ----------
    graph : input graph

    v : vertex
    """
    return sorted(graph.neighbors(v))


#
# In-house parsers
#


# kth graph format reader
def _read_graph_kthlist_format(inputfile,
                               graph_class=networkx.DiGraph,
                               bipartition=False):
    """Read a graph from file, in the KTH reverse adjacency lists format.

    If the vertices are listed from to sources to the sinks, then the
    graph is marked as topologically sorted, and any DAG test will be
    answered without running any visit to the graph. Otherwise no
    assumption is made.

    If the file is supposed to represent a bipartite graph (i.e.
    when `bipartition` parameters is true) then each target node of an
    edge will be part of the left side and and each source node will
    be on the right side. Vertices that are never named are considered
    to be isolated vertices on the right side.

    Parameters
    ----------
    inputfile : file object
        file handle of the input

    graph_class: class object
        the graph class to read, one of networkx.DiGraph (default)
        networkx.MultiDiGraph networkx.Graph networkx.MultiGraph

    bipartition : boolean
        enforce that each edge (u,v) is such that u in in the left size

    Raises
    ------
    ValueError
        Error parsing the file
    """
    if graph_class not in [
            networkx.DiGraph, networkx.MultiDiGraph, networkx.Graph,
            networkx.MultiGraph
    ]:
        raise RuntimeError(
            "[Internal error] Attempt to use an unsupported class for graph representation."
        )

    G = graph_class()
    G.name = ''
    G.ordered_vertices = []

    # is the input topologically sorted?
    topologically_sorted_input = True

    # vertex number
    nvertex = -1
    vertex_cnt = -1

    for i, l in enumerate(inputfile.readlines()):

        # add the comment to the header
        if l[0] == 'c':
            G.name += l[2:]
            continue

        # empty line
        if len(l.strip()) == 0:
            continue

        if ':' not in l:
            # vertex number spec
            if nvertex >= 0:
                raise ValueError(
                    "Line {} contains a second spec directive.".format(i))
            try:
                nvertex = int(l.strip())
                if nvertex < 0:
                    raise ValueError
            except ValueError:
                raise ValueError(
                    "Non negative number expected at line {}.".format(i))
            G.add_nodes_from(range(1, nvertex + 1))
            G.ordered_vertices = range(1, nvertex + 1)
            continue

        # Load edges from this line
        target, sources = l.split(':')
        try:
            target = int(target.strip())
            sources = [int(s) for s in sources.split()]
        except ValueError:
            raise ValueError("Non integer vertex ID at line {}.".format(i))
        if len(sources) < 1 or sources[-1] != 0:
            raise ValueError("Line {} must end with 0.".format(i))

        if target < 1 or target > nvertex:
            raise ValueError(
                "Vertex ID out of range [1,{}] at line {}.".format(nvertex, i))

        sources.pop()
        if len([x for x in sources if x < 1 or x > nvertex]) > 0:
            raise ValueError(
                "Vertex ID out of range [1,{}] at line {}.".format(nvertex, i))

        # Vertices should appear in increasing order if the graph is topologically sorted
        for s in sources:
            if s <= target:
                topologically_sorted_input = False

        # Check the bi-coloring on both side
        if bipartition:
            colors = [
                G.nodes[s]['bipartite'] for s in sources
                if 'bipartite' in G.nodes[s]
            ]

            if 'bipartite' in G.nodes[target]:
                colors += [1 - G.nodes[target]['bipartite']]

            if len(set(colors)) > 1:
                raise ValueError(
                    "Greedy bicoloring incompatible with edges in line {}.".
                    format(i))

            default_color = 0 if len(colors) == 0 else 1 - colors[0]
            G.nodes[target]['bipartite'] = default_color
            for s in sources:
                G.nodes[s]['bipartite'] = 1 - default_color

        # after vertices, add the edges
        for s in sources:
            G.add_edge(s, target)

    # label the bipartition on residual vertices
    if bipartition:
        for v in G.ordered_vertices:
            if 'bipartite' not in G.nodes[v]:
                G.nodes[v]['bipartite'] = 1

    # cache the information that the graph is topologically sorted.
    if topologically_sorted_input:
        G.topologically_sorted = True

    if nvertex != G.order():
        raise ValueError("{} vertices expected. Got {} instead.".format(
            nvertex, G.order()))
    return G


def _read_graph_dimacs_format(inputfile, graph_class=networkx.Graph):
    """Read a graph simple from file, in the DIMACS edge format.

    Parameters
    ----------
    inputfile : file object
        file handle of the input

    graph_class: class object
        the graph class to read, one of networkx.DiGraph (default)
        networkx.MultiDiGraph networkx.Graph networkx.MultiGraph
    """
    if not graph_class in [
            networkx.Graph, networkx.MultiGraph, networkx.DiGraph,
            networkx.MultiDiGraph
    ]:
        raise ValueError(
            "[Internal error] Attempt to use an unsupported class for graph representation."
        )

    G = graph_class()
    G.name = ''

    n = -1
    m = -1
    m_cnt = 0

    # is the input topologically sorted?
    for i, l in enumerate(inputfile.readlines()):

        # add the comment to the header
        if l[0] == 'c':
            G.name += l[2:]
            continue

        # parse spec line
        if l[0] == 'p':
            if n >= 0:
                raise ValueError(
                    "[Syntax error] " +
                    "Line {} contains a second spec line.".format(i))
            _, fmt, nstr, mstr = l.split()
            if fmt != 'edge':
                raise ValueError("[Input error] " +
                                 "Dimacs \'edge\' format expected.")
            n = int(nstr)
            m = int(mstr)
            G.add_nodes_from(range(1, n + 1))
            continue

        # parse spec line
        if l[0] == 'e':
            m_cnt += 1
            _, v, w = l.split()
            G.add_edge(int(v), int(w))

    if m != m_cnt:
        raise ValueError("[Syntax error] " +
                         "{} edges were expected.".format(m))

    return G


def _read_graph_matrix_format(inputfile):
    """Read a bipartite graph from file, in the adjiacency matrix format.

    This is an example of an adjacency matrix for a bipartite graph
    with 9 vertices on one side and 15 on the another side.

    .. 9 15
       1 1 0 1 0 0 0 1 0 0 0 0 0 0 0
       0 1 1 0 1 0 0 0 1 0 0 0 0 0 0
       0 0 1 1 0 1 0 0 0 1 0 0 0 0 0
       0 0 0 1 1 0 1 0 0 0 1 0 0 0 0
       0 0 0 0 1 1 0 1 0 0 0 1 0 0 0
       0 0 0 0 0 1 1 0 1 0 0 0 1 0 0
       0 0 0 0 0 0 1 1 0 1 0 0 0 1 0
       0 0 0 0 0 0 0 1 1 0 1 0 0 0 1
       1 0 0 0 0 0 0 0 1 1 0 1 0 0 0

    Parameters
    ----------
    inputfile: file object
        the file containing the graph specification

    Returns
    -------
    G : networkx.Graph

    """
    G = networkx.Graph()
    G.name = ''

    def scan_integer(inputfile):

        num_buffer = []
        line_cnt = 0

        while True:
            if len(num_buffer) == 0:

                line = inputfile.readline()

                if len(line) == 0:
                    raise StopIteration  # end of file

                line_cnt += 1
                tokens = line.split()

                if len(tokens) == 0 or tokens[0][0] == '#':
                    continue  # comment line

                try:
                    num_buffer.extend((int(lit), line_cnt) for lit in tokens)
                except ValueError:
                    raise ValueError("[Syntax error] " +
                                     "Line {} contains a non numeric entry.".
                                     format(line_cnt))

            yield num_buffer.pop(0)

    scanner = scan_integer(inputfile)

    try:
        n = next(scanner)[0]
        m = next(scanner)[0]

        # bipartition of vertices
        for i in range(1, n + 1):
            G.add_node(i, bipartite=0)
        for i in range(n + 1, n + m + 1):
            G.add_node(i, bipartite=1)

        # read edges
        for i in range(1, n + 1):
            for j in range(n + 1, n + m + 1):

                (b, l) = next(scanner)
                if b == 1:
                    G.add_edge(i, j)
                elif b == 0:
                    pass
                else:
                    raise ValueError(
                        "[Input error at line {}] Only 0 or 1 are allowed".
                        format(l))
    except StopIteration:
        raise ValueError("[Input error] Unexpected end of the matrix")

    # check that there are is no more data
    try:
        (b, l) = next(scanner)
        raise ValueError(
            "[Input error at line {}] There are more than {}x{} entries".
            format(l, n, m))
    except StopIteration:
        pass

    assert has_bipartition(G)
    return G


#
# In-house graph writers
#
def _write_graph_kthlist_format(G, output_file, bipartition=False):
    """Wrire a graph to a file, in the KTH reverse adjacency lists format.

    Parameters
    ----------
    G : graph object

    output_file : file object
        file handle of the output

    bipartition : boolean
        only write the left side adjacency lists. This will enforce
        the bipartition where the file is read again
    """

    print("c {}".format(G.name), file=output_file)
    print("{}".format(G.order()), file=output_file)

    # we need numerical indices for the vertices
    # adj list in the same order
    indices = {v: i for (i, v) in enumerate(enumerate_vertices(G), start=1)}

    if bipartition:
        V, _ = bipartite_sets(G)
    else:
        V = enumerate_vertices(G)

    from io import StringIO
    output = StringIO()

    for v in V:

        if G.is_directed():
            nbors = [indices[w] for w in G.predecessors(v)]
        else:
            nbors = [indices[w] for w in G.adj[v].keys()]

        nbors.sort()

        output.write(str(indices[v]) + " : ")
        output.write(" ".join([str(i) for i in nbors]))
        output.write(" 0\n")

    print(output.getvalue(), file=output_file)


def _write_graph_dimacs_format(G, output_file):
    """Wrire a graph to a file, in DIMACS format.

    Parameters
    ----------
    G : graph object

    output_file : file object
        file handle of the output
    """

    print("c {}".format(G.name).strip(), file=output_file)
    vertices = dict((name, index) for index, name in enumerate(G.nodes(), 1))
    edges = G.edges()
    print("p edge {} {}".format(len(vertices), len(edges)), file=output_file)

    for v, w in edges:
        print("e {} {}".format(vertices[v], vertices[w]), file=output_file)


def _write_graph_matrix_format(G, output_file):
    """Wrire a graph to a file, in \"matrix\" format.

    Parameters
    ----------
    G : graph object

    output_file : file object
        file handle of the output
    """

    Left, Right = bipartite_sets(G)

    print("{} {}".format(len(Left), len(Right)), file=output_file)
    for u in Left:

        adj_row = []

        for v in Right:
            if G.has_edge(u, v):
                adj_row.append("1")
            else:
                adj_row.append("0")

        print(" ".join(adj_row), file=output_file)


#
# Graph generator (missing from networkx)
#


def bipartite_random_left_regular(l, r, d, seed=None):
    """Returns a random bipartite graph with constant left degree.

    Each vertex on the left side has `d` neighbors on the right side,
    picked uniformly at random without repetition.

    Each vertex in the graph has an attribute `bipartite` which is 0
    for the vertices on the left side and 1 for the vertices on the
    right side.

    Parameters
    ----------
    l : int
        vertices on the left side
    r : int
        vertices on the right side
    d : int
        degree on the left side.
    seed : hashable object
        seed the random generator

    Returns
    -------
    networkx.Graph

    Raises
    ------
    ValueError
        unless ``l``, ``r`` and ``d`` are non negative.

    """
    import random
    if seed:
        random.seed(seed)

    if l < 0 or r < 0 or d < 0:
        raise ValueError(
            "bipartite_random_left_regular(l,r,d) needs l,r,d >=0.")

    G = networkx.Graph()
    G.name = "bipartite_random_left_regular({},{},{})".format(l, r, d)

    L = list(range(0, l))
    R = list(range(l, l + r))
    d = min(r, d)

    for u in L:
        G.add_node(u, bipartite=0)

    for v in R:
        G.add_node(v, bipartite=1)

    for u in L:
        for v in sorted(random.sample(R, d)):
            G.add_edge(u, v)

    return G


def bipartite_shift(N, M, pattern=[]):
    """Returns a bipartite graph where edges are a fixed shifted sequence.

    The graph has :math:`N` vertices on the left (numbered from
    :math:`1` to :math:`N`), and :math:`M` vertices on the right
    (numbered from :math:`1` to :math:`M`),


    The vertex :math:`1` on the left side has edges to vertices
    :math:`v_1`, :math:`v_2`, :math:`v_3`,... (mod :math:`M`).
    Each other vertex :math:`i>1` on the left side has edges
    :math:`i+v_1`, :math:`i+v_2`, :math:`i+v_3`,... with wrap around
    over :math:`[1..M]`.

    Parameters
    ----------
    N : int
        vertices on the left side
    M : int
        vertices on the right side
    pattern : list(int)
        pattern of neighbors

    Returns
    -------
    networkx.Graph

    Raises
    ------
    ValueError
        unless ``N``, ``M`` are non negative and ``pattern`` has vertices outside the range.

    """
    if N < 1 or M < 1:
        raise ValueError("bipartite_shift(N,M,pattern) needs N,M >= 0.")

    if any([x < 1 or x > M for x in pattern]):
        raise ValueError(
            "bipartite_shift(N,M,pattern) needs 1 <= pattern[i] <= M.")

    G = networkx.Graph()
    G.name = "bipartite_shift_regular({},{},{})".format(N, M, pattern)

    L = list(range(1, N + 1))
    R = list(range(N + 1, N + M + 1))

    for u in L:
        G.add_node(u, bipartite=0)

    for v in R:
        G.add_node(v, bipartite=1)

    pattern.sort()
    for u in L:
        for i in pattern:
            G.add_edge(u, N + 1 + (i + u - 2) % M)

    return G


def bipartite_random_regular(l, r, d, seed=None):
    """Returns a random bipartite graph with constant degree on both sides.

    The graph is d-regular on the left side and regular on the right
    size, so it must be that d*l / r is an integer number.

    Each vertex in the graph has an attribute `bipartite` which is 0
    for the vertices on the left side and 1 for the vertices on the
    right side.

    Parameters
    ----------
    l : int
       vertices on the left side
    r : int
       vertices on the right side
    d : int
       degree of vertices at the left side
    seed : hashable object
       seed of random generator

    Returns
    -------
    networkx.Graph

    Raises
    ------
    ValueError
        if one among ``l``, ``r`` and ``d`` is negative or
        if ``r`` does not divides `l*d`

    References
    ----------
    [1] http://...

    """

    import random
    if seed:
        random.seed(seed)

    if l < 0 or r < 0 or d < 0:
        raise ValueError("bipartite_random_regular(l,r,d) needs l,r,d >=0.")

    if (l * d) % r != 0:
        raise ValueError(
            "bipartite_random_regular(l,r,d) needs r to divid l*d.")

    G = networkx.Graph()
    G.name = "bipartite_random_regular({},{},{})".format(l, r, d)

    L = list(range(0, l))
    R = list(range(l, l + r))

    for u in L:
        G.add_node(u, bipartite=0)

    for v in R:
        G.add_node(v, bipartite=1)

    A = L * d
    B = R * (l * d // r)
    assert len(B) == l * d

    for i in range(l * d):
        # Sample an edge, do not add it if it existed
        # We expect to sample at most d^2 edges
        for retries in range(3 * d * d):
            ea = random.randint(i, l * d - 1)
            eb = random.randint(i, l * d - 1)
            if not G.has_edge(A[ea], B[eb]):
                G.add_edge(A[ea], B[eb])
                A[i], A[ea] = A[ea], A[i]
                B[i], B[eb] = B[eb], B[i]
                break
        else:
            # Sampling takes too long, maybe no good edge exists
            failure = True
            for ea in range(i, l * d):
                for eb in range(i, l * d):
                    if not G.has_edge(A[ea], B[eb]):
                        failure = False
                        break
                if not failure:
                    break
            if failure:
                return bipartite_random_regular(l, r, d)

    return G


def dag_pyramid(height):
    """Generates the pyramid DAG

    Parameters
    ----------
    height : int
        the height of the tree

    Returns
    -------
    networkx.DiGraph
    """
    D = networkx.DiGraph()
    D.name = 'Pyramid of height {}'.format(height)
    D.ordered_vertices = []

    # vertices
    X=[ [('x_{{{},{}}}'.format(h,i),h,i) for i in range(height-h+1)] \
        for h in range(height+1) ]

    for layer in X:
        for (name, h, i) in layer:
            D.add_node(name, rank=(h, i))
            D.ordered_vertices.append(name)

    # edges
    for h in range(1, len(X)):
        for i in range(len(X[h])):
            D.add_edge(X[h - 1][i][0], X[h][i][0])
            D.add_edge(X[h - 1][i + 1][0], X[h][i][0])

    return D


def dag_complete_binary_tree(height):
    """Generates the complete binary tree DAG

    Parameters
    ----------
    height : int
        the height of the tree

    Returns
    -------
    networkx.DiGraph
    """
    D = networkx.DiGraph()
    D.name = 'Complete binary tree of height {}'.format(height)
    D.ordered_vertices = []
    # vertices
    vert = ['v_{}'.format(i) for i in range(1, 2 * (2**height))]
    for w in vert:
        D.add_node(w)
        D.ordered_vertices.append(w)
    # edges
    N = len(vert) - 1
    for i in range(len(vert) // 2):
        D.add_edge(vert[N - 2 * i - 1], vert[N - i])
        D.add_edge(vert[N - 2 * i - 2], vert[N - i])

    return D


def sample_missing_edges(G, m, seed=None):
    """Sample m pairs of missing edges in G

    If :math:`G` is not complete and has at least :math:`m` missing edges, :math:`m` of them are sampled.

    Parameters
    ----------
    G : networkx.Graph
        a graph with at least :math:`m` missing edges
    m : int
       the number of missing edges to sample
    seed : hashable object
       seed of random generator

    Returns
    -------
    list of edges

    Raises
    ------
    ValueError
        if :math:`G` doesn't have :math:`m` missing edges
    RuntimeError
        Sampling failure in the sparse case


    """

    import random
    if seed:
        random.seed(seed)

    from networkx import non_edges

    if m < 0:
        raise ValueError("You can only sample a non negative number of edges.")

    total_number_of_edges = None

    if has_bipartition(G):

        Left, Right = bipartite_sets(G)
        total_number_of_edges = len(Left) * len(Right)

        def edge_sampler():
            u = random.sample(Left, 1)[0]
            v = random.sample(Right, 1)[0]
            return (u, v)

        def available_edges():
            return [(u, v) for u in Left for v in Right
                    if not G.has_edge(u, v)]

    else:

        total_number_of_edges = G.order() * (G.order() - 1) / 2

        def edge_sampler():
            return random.sample(G.nodes(), 2)

        def available_edges():
            return list(non_edges(G))

    number_avaiable_edges = total_number_of_edges - G.number_of_edges()

    if number_avaiable_edges < m:
        raise ValueError(
            "The graph does not have {} missing edges to sample.".format(m))

    if G.number_of_edges() + m >= total_number_of_edges / 2:
        # Large density case: enumerate missing edges and sample.
        return random.sample(available_edges(), m)

    else:
        # Sparse case: sample and retry
        missing_edges = set()

        for _ in range(100 * m):

            if len(missing_edges) >= m:
                break

            u, v = edge_sampler()
            if (u,v) not in missing_edges and \
               (v,u) not in missing_edges and \
               not G.has_edge(u,v):
                missing_edges.add((u, v))

        if len(missing_edges) >= m:
            return missing_edges
        else:
            raise RuntimeError(
                "Improbable failure at sampling missing edges in a sparse graph."
            )



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Formulas that encode coloring related problems
"""

from cnfformula.cnf import CNF
from cnfformula.graphs import enumerate_vertices
from cnfformula.graphs import enumerate_edges
from cnfformula.graphs import neighbors

from itertools import combinations
import collections


def GraphColoringFormula(G, colors, functional=True):
    """Generates the clauses for colorability formula

    The formula encodes the fact that the graph :math:`G` has a coloring
    with color set ``colors``. This means that it is possible to
    assign one among the elements in ``colors``to that each vertex of
    the graph such that no two adjacent vertices get the same color.

    Parameters
    ----------
    G : networkx.Graph
        a simple undirected graph
    colors : list or positive int
        a list of colors or a number of colors

    Returns
    -------
    CNF
       the CNF encoding of the coloring problem on graph ``G``

    """
    if isinstance(colors, int) and colors >= 0:
        colors = list(range(1, colors + 1))

    if not isinstance(colors, collections.Iterable):
        ValueError(
            "Parameter \"colors\" is expected to be a number or an iterable")

    # Describe the formula
    name = "Graph {}-Colorability".format(len(colors))

    if hasattr(G, 'name'):
        header = name + " of graph: " + G.name
    col = CNF(description=header)

    # Fix the vertex order
    V = enumerate_vertices(G)

    # Each vertex has a color
    for vertex in V:
        clause = []
        for color in colors:
            clause += [(True, 'x_{{{0},{1}}}'.format(vertex, color))]
        col.add_clause(clause)

        # unique color per vertex
        if functional:
            for (c1, c2) in combinations(colors, 2):
                col.add_clause([(False, 'x_{{{0},{1}}}'.format(vertex, c1)),
                                (False, 'x_{{{0},{1}}}'.format(vertex, c2))],
                               strict=True)

    # This is a legal coloring
    for (v1, v2) in enumerate_edges(G):
        for c in colors:
            col.add_clause([(False, 'x_{{{0},{1}}}'.format(v1, c)),
                            (False, 'x_{{{0},{1}}}'.format(v2, c))],
                           strict=True)

    return col


def EvenColoringFormula(G):
    """Even coloring formula

    The formula is defined on a graph :math:`G` and claims that it is
    possible to split the edges of the graph in two parts, so that
    each vertex has an equal number of incident edges in each part.

    The formula is defined on graphs where all vertices have even
    degree. The formula is satisfiable only on those graphs with an
    even number of vertices in each connected component [1]_.

    Arguments
    ---------
    G : networkx.Graph 
       a simple undirected graph where all vertices have even degree

    Raises
    ------
    ValueError
       if the graph in input has a vertex with odd degree

    Returns
    -------
    CNF object

    References
    ----------
    .. [1] Locality and Hard SAT-instances, Klas Markstrom
       Journal on Satisfiability, Boolean Modeling and Computation 2 (2006) 221-228

    """
    description = "Even coloring formula on " + G.name
    F = CNF(description=description)

    def var_name(u, v):
        if u <= v:
            return 'x_{{{0},{1}}}'.format(u, v)
        else:
            return 'x_{{{0},{1}}}'.format(v, u)

    for (u, v) in enumerate_edges(G):
        F.add_variable(var_name(u, v))

    # Defined on both side
    for v in enumerate_vertices(G):

        if G.degree(v) % 2 == 1:
            raise ValueError(
                "Markstrom formulas requires all vertices to have even degree."
            )

        edge_vars = [var_name(u, v) for u in neighbors(G, v)]

        for cls in CNF.equal_to_constraint(edge_vars, len(edge_vars) // 2):
            F.add_clause(cls, strict=True)

    return F



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""CNF Formulas for Ramsey-like statements
"""

from cnfformula.cnf import CNF

from textwrap import dedent
from itertools import combinations
from math import sqrt


def PythagoreanTriples(N):
    """There is a Pythagorean triples free coloring on N 
    
    The formula claims that it is possible to bicolor the numbers from
    1 to :math:`N` so that there  is no monochromatic triplet 
    :math:`(x,y,z)` so that :math:`x^2+y^2=z^2`.

    Parameters
    ----------
    N  : int
         size of the interval

    Return
    ------
    A CNF object

    Raises
    ------
    ValueError
       Parameters are not positive integers

    References
    ----------
    .. [1] M. J. Heule, O. Kullmann, and V. W. Marek. 
           Solving and verifying the boolean pythagorean triples problem via cube-and-conquer. 
           arXiv preprint arXiv:1605.00723, 2016.
    """

    description = "Pythagorean triples problem on 1...{}".format(N)
    ptn = CNF(description=description)

    def V(i):
        return "x_{{{}}}".format(i)

    # Variables represent the coloring of the number
    for i in range(1, N + 1):
        ptn.add_variable(V(i))

    for x, y in combinations(range(1, N + 1), 2):
        z = int(sqrt(x**2 + y**2))
        if z <= N and z**2 == x**2 + y**2:
            ptn.add_clause([(True, V(x)), (True, V(y)), (True, V(z))],
                           strict=True)
            ptn.add_clause([(False, V(x)), (False, V(y)), (False, V(z))],
                           strict=True)

    return ptn


def RamseyLowerBoundFormula(s, k, N):
    """Formula claiming that Ramsey number r(s,k) > N

    Arguments:
    - `s`: independent set size
    - `k`: clique size
    - `N`: vertices
    """

    description = "{}-vertices graph free of {}-independent sets and {}-cliques".format(
        N, s, k)
    ram = CNF(description=description)

    def V(i):
        return "x_{{{}}}".format(i)

    #
    # One variable per edge (indices are ordered)
    #
    for edge in combinations(range(1, N + 1), 2):
        ram.add_variable('e_{{{0},{1}}}'.format(*edge))

    #
    # No independent set of size s
    #
    for vertex_set in combinations(range(1, N + 1), s):
        clause = []
        for edge in combinations(vertex_set, 2):
            clause += [(True, 'e_{{{0},{1}}}'.format(*edge))]
        ram.add_clause(clause, strict=True)

    #
    # No clique of size k
    #
    for vertex_set in combinations(range(1, N + 1), k):
        clause = []
        for edge in combinations(vertex_set, 2):
            clause += [(False, 'e_{{{0},{1}}}'.format(*edge))]
        ram.add_clause(clause, strict=True)

    return ram



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Command line tools based on cnfformula package

- main cnfgen implementation
- kthlist2pebbling
- cnfshuffle
- cnftransform
"""

from .kthlist2pebbling import cli as kthlist2pebbling
from .cnfshuffle import cli as cnfshuffle
from .main import cli as cnfgen

from .cmdline import CLIError



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of Thapen's size-width tradeoff formula
"""

from itertools import product

from cnfformula.cnf import CNF


def CPLSFormula(a, b, c):
    """Thapen's size-width tradeoff formula

    The formula is a propositional version of the coloured polynomial
    local search principle (CPLS). A description can be found in [1]_.

    Parameters
    ----------
    a: integer
       number of levels
    b: integer
       nodes per level (must be a power of 2)
    c: integer
       number of colours (must be a power of 2)

    References
    ----------
    .. [1] N. Thapen (2016)
           Trade-offs between length and width in resolution.
           Theory of Computing, 12(1), 1–14.

    """
    if a <= 0:
        raise ValueError("a must be positive.")
    if b <= 0 or b & (b - 1):
        raise ValueError("b must be a power of two.")
    if c <= 0 or c & (c - 1):
        raise ValueError("c must be a power of two.")

    log_b = 0
    while 2**log_b < b:
        log_b += 1
    log_c = 0
    while 2**log_c < c:
        log_c += 1

    description = "Thapen's CPLS formula with {} levels, {} nodes per level, {} colours".format(
        a, b, c)
    formula = CNF(description=description)

    # this allows notation G[i](x, y) and similar, which looks more like G_{i}(x, y) than G(i, x, y)
    class func_list(list):
        def __init__(self, f):
            self.f = f

        def __getitem__(self, key):
            return self.f(key)

    # TODO: Is "log a" a typo in "3." of the variables definition? I assume it's "log c".

    # G_i(x, y) -> G[i](x, y)
    G = func_list(lambda i: lambda x, y: 'G_{{{0}}}({1},{2})'.format(i, x, y))
    # (f_i(x))_j -> f[i](x)[j]
    f = func_list(lambda i: lambda x: func_list(
        lambda j: '(f_{{{0}}}({1}))_{{{2}}}'.format(i, x, j)))
    # (u(x))_j -> u(x)[j]
    u = lambda x: func_list(lambda j: '(u({0}))_{{{1}}}'.format(x, j))

    # 1. For each i < a, x < b and y < c, there is a variable G_i(x,y).
    for i, x, y in product(range(a), range(b), range(c)):
        formula.add_variable(G[i](x, y))
    # 2. For each i < a, x < b and j < log b, there is a variable (f_i(x))_j,
    #    standing for the j'th bit of the value of f_i(x).
    for i, x, j in product(range(a), range(b), range(log_b)):
        formula.add_variable(f[i](x)[j])
    # 3. For each x < b and j < log a, there is a variable (u(x))_j,
    #    standing for the j'th bit of the value of u(x).
    for x, j in product(range(b), range(log_c)):
        formula.add_variable(u(x)[j])

    # x_{bits-1}..x_0 != binary(x'), encoded as the disjunction of l_j for all j,
    # where l_j = ~x_j if the j'th bit of x' is 1, and l_j = x_j otherwise.
    def bin_ineq(f_i_x, x_prime, bits):
        return [(not bool(x_prime & (1 << j)), f_i_x[j]) for j in range(bits)]

    # 1. For each y < c, the clause ~G_0(0, y)
    for y in range(c):
        formula.add_clause([(False, G[0](0, y))], strict=True)
    # 2. For each i < a-1, each pair x,x' < b and each y < c,
    #    the clause f_i(x) = x' ^ G_{i+1}(x', y) -> G_i(x, y)
    for i, x, x_, y in product(range(a - 1), range(b), range(b), range(c)):
        formula.add_clause(bin_ineq(f[i](x), x_, log_b) +
                           [(False, G[i + 1](x_, y)), (True, G[i](x, y))],
                           strict=True)
    # 3. For each x < b and each y < c, the clause u(x) = y -> G_{a-1}(x, y)
    for x, y in product(range(b), range(c)):
        formula.add_clause(bin_ineq(u(x), y, log_c) + [(True, G[a - 1](x, y))],
                           strict=True)

    return formula



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of Tseitin formulas
"""

from cnfformula.cnf import CNF

from cnfformula.graphs import enumerate_vertices,neighbors

import random

def TseitinFormula(graph,charges=None):
    """Build a Tseitin formula based on the input graph.

    Odd charge is put on the first vertex by default, unless other
    vertices are is specified in input.

    Arguments:
    - `graph`: input graph
    - `charges': odd or even charge for each vertex
    """
    V=enumerate_vertices(graph)

    if charges==None:
        charges=[1]+[0]*(len(V)-1)             # odd charge on first vertex
    else:
        charges = [bool(c) for c in charges]   # map to boolean

    if len(charges)<len(V):
        charges=charges+[0]*(len(V)-len(charges))  # pad with even charges

    # init formula
    tse=CNF()
    edgename = { }
    
    for (u,v) in sorted(graph.edges(),key=sorted):
        edgename[(u,v)] =  "E_{{{0},{1}}}".format(u,v)
        edgename[(v,u)] =  "E_{{{0},{1}}}".format(u,v)
        tse.add_variable(edgename[(u,v)])

    # add constraints
    for v,c in zip(V,charges):
        
        # produce all clauses and save half of them
        names = [ edgename[(u,v)] for u in neighbors(graph,v) ]
        for cls in CNF.parity_constraint(names,c):
            tse.add_clause(list(cls),strict=True)

    return tse




#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Screen messages for CNFgen interaction
"""

import sys
import textwrap
from contextlib import contextmanager

_prefix = ''


@contextmanager
def msg_prefix(new_prefix=''):
    """Set prefix for all interactive messages"""
    global _prefix
    old_prefix = _prefix
    _prefix = old_prefix + new_prefix
    yield
    _prefix = old_prefix


def interactive_msg(msg, filltext=None):
    """Writes a message to the interactive user (if present).

    When the input comes from an interactive user on the terminal, it
    is useful to give them feedback regarding the expected output.
    This message is sent to stderr in order to help interactive usage,
    but it is not sent out if input comes from a non interactive
    terminal.
    """
    global _prefix
    msg = textwrap.dedent(msg)
    if filltext is not None and filltext > len(_prefix) + 30:
        msg = textwrap.fill(msg, width=filltext - len(_prefix))
    msg = textwrap.indent(msg, _prefix, lambda line: True)

    if sys.stdin.isatty():
        print(msg, file=sys.stderr)


def error_msg(msg, filltext=None):
    """Writes an error message.

    """
    global _prefix
    msg = textwrap.dedent(str(msg))
    if filltext is not None and filltext > 0:
        msg = textwrap.fill(msg, width=filltext - len(_prefix))
    msg = textwrap.indent(msg, _prefix, lambda line: True)
    print(msg, file=sys.stderr)


class InternalBug(Exception):
    """Bug related to internal consistency
    """
    def __init__(self, msg):

        bug_msg = """INTERNAL ERROR

{}

Ooops! This was never supposed to happen. Please
take note of your command line and send it to
<massimo.lauria@uniroma1.it>
""".format(str(msg))
        super(Exception, self).__init__(bug_msg)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Formula Helper interface

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""


class FormulaHelper:
    """Command line helper for a formula family"""

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line parser for this formula subcommand"""
        raise NotImplementedError("Formula family helper must be subclassed")

    @staticmethod
    def build_cnf(args):
        """Buil the CNF according to the parameters on the command line"""
        raise NotImplementedError("Formula family helper must be subclassed")





#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of counting/matching formulas
"""

from cnfformula.cnf import CNF
from cnfformula.graphs import enumerate_vertices
from cnfformula.graphs import neighbors

from itertools import combinations


def CountingPrinciple(M, p):
    """Generates the clauses for the counting matching principle.
    
    The principle claims that there is a way to partition M in sets of
    size p each.

    Arguments:
    - `M`  : size of the domain
    - `p`  : size of each class

    """
    if M < p:
        raise ValueError(
            "Domain size must be larger or equal than the class size")
    description = "Counting Principle: {0} divided in parts of size {1}.".format(
        M, p)
    cnf = CNF(description=description)

    def var_name(tpl):
        return "Y_{{" + ",".join("{0}".format(v) for v in tpl) + "}}"

    # Incidence lists
    incidence = [[] for _ in range(M)]
    for tpl in combinations(range(M), p):
        for i in tpl:
            incidence[i].append(tpl)

    # Each element of the domain is in exactly one part.
    for el in range(M):

        edge_vars = [var_name(tpl) for tpl in incidence[el]]

        for cls in CNF.equal_to_constraint(edge_vars, 1):
            cnf.add_clause(cls)

    return cnf


def PerfectMatchingPrinciple(G):
    """Generates the clauses for the graph perfect matching principle.
    
    The principle claims that there is a way to select edges to such
    that all vertices have exactly one incident edge set to 1.

    Parameters
    ----------
    G : undirected graph

    """
    # Describe the formula
    description = "Perfect Matching Principle"

    if hasattr(G, 'name'):
        description += " on " + G.name

    cnf = CNF(description=description)

    def var_name(u, v):
        if u <= v:
            return 'x_{{{0},{1}}}'.format(u, v)
        else:
            return 'x_{{{0},{1}}}'.format(v, u)

    # Each vertex has exactly one edge set to one.
    for v in enumerate_vertices(G):

        edge_vars = [var_name(u, v) for u in neighbors(G, v)]

        for cls in CNF.equal_to_constraint(edge_vars, 1):
            cnf.add_clause(cls)

    return cnf



#!/usr/bin/env python
copyright_years = '(C) 2012-2020'

__all__ = ['info']


def get_version():
    """Compute CNFgen verison number

Version number is obtained from the GIT repository, either
hardcoding that number into version.py file (for release) or
directly calling GIT (for development).
"""
    try:
        from cnfformula.version import version
        return version
    except ImportError:
        pass

    try:
        import subprocess
        version = subprocess.check_output(
            ["git", "describe", "--tags", "--always"]).strip().decode('utf-8')
        return version
    except subprocess.CalledProcessError:
        pass


# Project info
info = {
    'project': 'CNFgen',
    'description': 'CNF formula generator',
    'author': 'Massimo Lauria',
    'author_email': 'massimo.lauria@uniroma1.it',
    'version': 'unknown version',
    'license': 'GPL-3',
    'url': 'https://massimolauria.net/cnfgen',
    'repository': 'https://github.com/MassimoLauria/cnfgen'
}

# Setup additional info
info['copyright'] = "{0} {1} <{2}>".format(copyright_years, info['author'],
                                           info['author_email'])
info['version'] = get_version()



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Various utilities for the manipulation of the CNFs.
"""

import sys
from ..cnf import CNF


def dimacs2compressed_clauses(infile):
    """
    Parse a dimacs cnf file into a list of
    compressed clauses.

    return: (n,c) where

    n is the number of variables
    c is the list of compressed clauses.
    """
    n = -1  # negative signal that spec line has not been read
    m = -1

    my_header = ""
    my_clauses = []

    line_counter = 0
    literal_buffer = []

    for l in infile.readlines():

        line_counter += 1

        # Ignore comments
        if l[0] == 'c':
            continue

        # parse spec line
        if l[0] == 'p':
            if n >= 0:
                raise ValueError(
                    "There is a another spec at line {}".format(line_counter))
            try:
                _, _, nstr, mstr = l.split()
                n = int(nstr)
                m = int(mstr)
                l.split()
            except ValueError:
                raise ValueError("Spec at line {} should have "
                                 "format 'p cnf <n> <m>'".format(line_counter))
            continue

        # parse literals
        try:
            for lv in [int(lit) for lit in l.split()]:
                if lv == 0:
                    my_clauses.append(tuple(literal_buffer))
                    literal_buffer = []
                else:
                    literal_buffer.append(lv)
        except ValueError:
            raise ValueError("Invalid literal at line {}".format(line_counter))

    # Checks at the end of parsing
    if len(literal_buffer) > 0:
        raise ValueError("Last clause was incomplete")

    if m == '-1':
        raise ValueError("Missing spec line 'p cnf <n> <m>")

    if m != len(my_clauses):
        raise ValueError("Formula contains {} clauses "
                         "but {} were expected.".format(len(my_clauses), m))

    # return the formula
    return (n, my_clauses)


def readCNF(infile=None):
    """Read dimacs file into a CNF object

    By default it reads the file from standard input.
    """
    if infile is None:
        infile = sys.stdin
    if not hasattr(infile, 'name'):
        infile.name = 'unknown'

    nvariables, clauses = dimacs2compressed_clauses(infile)

    cnf = CNF(
        description='Formula read from DIMACS file {}'.format(infile.name))

    for i in range(1, nvariables + 1):
        cnf.add_variable(i)

    cnf._add_compressed_clauses(clauses)

    # return the formula
    cnf._check_coherence(force=True)
    return cnf



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Formula Helpers for pigeonhole principle formulas

Copyright (C) 2012-2020 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

from cnfformula import CliqueColoring
from cnfformula import RamseyLowerBoundFormula
from cnfformula import PythagoreanTriples

from cnfformula import PigeonholePrinciple
from cnfformula import GraphPigeonholePrinciple
from cnfformula import BinaryPigeonholePrinciple

from cnfformula.graphs import bipartite_random_left_regular

from .graph_cmdline import BipartiteGraphHelper
from .formula_helpers import FormulaHelper
from .cmdline import positive_int, nonnegative_int

import argparse

usage_string = """Pigeonhole Principle

Pigeonhole principle claims that P pigeons can fly to H holes with no
two pigeons in the same hole. This is unsatisfiable when P > H. It is
possible to specify a bipartite graph that specifies which pigeons can
fly to which holes.

positional arguments:
 {0} N           --- N+1 pigeons fly to N holes
 {0} M N         --- M pigeons fly to N holes
 {0} M N D       --- M pigeons fly to N holes, pigeon left degree D
 """

example_string = """examples:
 {0} 100           --- 101 pigeons and 100 holes (unsat)
 {0} 14 10         --- 14 pigeons and 10 holes (unsat)
 {0} 9  10         --- 9  pigeons and 10 holes (sat)
 {0} 12 10 3       --- 12 pigeons and 10 holes,
 {1}                   pigeon can go to 3 random holes
 """


class PHPArgs(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        if len(values) == 0:
            parser.error('php formula needs <pigeons> <holes> specification')

        if len(values) > 3:
            parser.error('too many arguments for php formula')

        try:
            intvalues = [int(x) for x in values]
            if not all([x >= 0 for x in intvalues]):
                raise ValueError
        except ValueError:
            parser.error('php arguments must be positive integers')

        if len(intvalues) == 1:
            setattr(args, "pigeons", intvalues[0] + 1)
            setattr(args, "holes", intvalues[0])
            setattr(args, "degree", intvalues[0])
        elif len(intvalues) == 2:
            setattr(args, "pigeons", intvalues[0])
            setattr(args, "holes", intvalues[1])
            setattr(args, "degree", intvalues[1])
        else:
            setattr(args, "pigeons", intvalues[0])
            setattr(args, "holes", intvalues[1])
            setattr(args, "degree", intvalues[2])

            if intvalues[1] < intvalues[2]:
                parser.error("degree must be at most the number of holes")


class PHPCmdHelper(FormulaHelper):
    """Command line helper for the Pigeonhole principle CNF"""

    name = 'php'
    description = 'pigeonhole principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for pigeonhole principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.usage = usage_string.format(parser.prog)
        parser.description = example_string.format(parser.prog,
                                                   " " * len(parser.prog))
        parser.add_argument('pigeonholes',
                            action=PHPArgs,
                            nargs='*',
                            help=argparse.SUPPRESS)
        parser.add_argument('--functional',
                            action='store_true',
                            help="pigeons sit in at most one hole")
        parser.add_argument('--onto',
                            action='store_true',
                            help="every hole has a sitting pigeon")

    @staticmethod
    def build_cnf(args):
        """Build a PHP formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        if args.holes == args.degree:
            return PigeonholePrinciple(args.pigeons,
                                       args.holes,
                                       functional=args.functional,
                                       onto=args.onto)
        else:
            G = bipartite_random_left_regular(args.pigeons, args.holes,
                                              args.degree)
            return GraphPigeonholePrinciple(G,
                                            functional=args.functional,
                                            onto=args.onto)


class GPHPCmdHelper(FormulaHelper):
    """Command line helper for the Pigeonhole principle on graphs"""

    name = 'gphp'
    description = 'graph pigeonhole principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for pigeonhole principle formula over graphs

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('--functional',
                            action='store_true',
                            help="pigeons sit in at most one hole")
        parser.add_argument('--onto',
                            action='store_true',
                            help="every hole has a sitting pigeon")
        BipartiteGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build a Graph PHP formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = BipartiteGraphHelper.obtain_graph(args)
        return GraphPigeonholePrinciple(G,
                                        functional=args.functional,
                                        onto=args.onto)


class BPHPCmdHelper(FormulaHelper):
    """Command line helper for the Pigeonhole principle CNF"""

    name = 'bphp'
    description = 'binary pigeonhole principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for pigeonhole principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('pigeons',
                            metavar='<pigeons>',
                            type=positive_int,
                            help="Number of pigeons")
        parser.add_argument('holes',
                            metavar='<holes>',
                            type=positive_int,
                            help="Number of holes")

    @staticmethod
    def build_cnf(args):
        """Build a PHP formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        return BinaryPigeonholePrinciple(args.pigeons, args.holes)


class CliqueColoringCmdHelper(FormulaHelper):
    """Command line helper for the Clique-coclique CNF"""

    name = 'cliquecoloring'
    description = 'There is a graph G with a k-clique and a c-coloring'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for clique-coloring formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('n',
                            metavar='<n>',
                            type=nonnegative_int,
                            help="Number of vertices")
        parser.add_argument('k',
                            metavar='<k>',
                            type=positive_int,
                            help="Clique size")
        parser.add_argument('c',
                            metavar='<c>',
                            type=positive_int,
                            help="Coloring size")

    @staticmethod
    def build_cnf(args):
        """Build a Clique-coclique formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        return CliqueColoring(args.n, args.k, args.c)


class RamseyCmdHelper(FormulaHelper):
    """Command line helper for RamseyNumber formulas
    """
    name = 'ram'
    description = 'ramsey number principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Ramsey formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('s',
                            metavar='<s>',
                            type=positive_int,
                            help="Forbidden independent set size")
        parser.add_argument('k',
                            metavar='<k>',
                            type=positive_int,
                            help="Forbidden independent clique")
        parser.add_argument('N',
                            metavar='<N>',
                            type=nonnegative_int,
                            help="Graph size")

    @staticmethod
    def build_cnf(args):
        """Build a Ramsey formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        return RamseyLowerBoundFormula(args.s, args.k, args.N)


class PTNCmdHelper(FormulaHelper):
    """Command line helper for PTN formulas
    """
    name = 'ptn'
    description = 'Bicoloring of N with no monochromatic Pythagorean Triples'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for PTN formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('N',
                            metavar='<N>',
                            type=nonnegative_int,
                            help="Size of the domain")

    @staticmethod
    def build_cnf(args):
        """Build a Ramsey formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        return PythagoreanTriples(args.N)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of counting/matching formulas helpers

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019 Massimo Lauria <massimo.lauria@uniroma1.it>
https://massimolauria.net/cnfgen/
"""

from cnfformula import CountingPrinciple
from cnfformula import PerfectMatchingPrinciple
from cnfformula import TseitinFormula
from cnfformula import SubsetCardinalityFormula

from .graph_cmdline import SimpleGraphHelper
from .graph_cmdline import BipartiteGraphHelper

from .formula_helpers import FormulaHelper

import random


class ParityCmdHelper(FormulaHelper):
    """Command line helper for Parity Principle formulas
    """
    name = 'parity'
    description = 'parity principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Parity Principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('N', metavar='<N>', type=int, help="domain size")

    @staticmethod
    def build_cnf(args):
        return CountingPrinciple(args.N, 2)


class PMatchingCmdHelper(FormulaHelper):
    """Command line helper for Perfect Matching Principle formulas
    """
    name = 'matching'
    description = 'perfect matching principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Perfect Matching Principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        G = SimpleGraphHelper.obtain_graph(args)
        return PerfectMatchingPrinciple(G)


class CountingCmdHelper(FormulaHelper):
    """Command line helper for Counting Principle formulas
    """
    name = 'count'
    description = 'counting principle'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Counting Principle formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument('M', metavar='<M>', type=int, help="domain size")
        parser.add_argument('p',
                            metavar='<p>',
                            type=int,
                            help="size of the parts")

    @staticmethod
    def build_cnf(args):
        """Build an Counting Principle formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        return CountingPrinciple(args.M, args.p)


class TseitinCmdHelper(FormulaHelper):
    """Command line helper for Tseitin  formulas
    """
    name = 'tseitin'
    description = 'tseitin formula'

    @staticmethod
    def setup_command_line(parser):
        """Setup the command line options for Tseitin formula

        Arguments:
        - `parser`: parser to load with options.
        """
        parser.add_argument(
            '--charge',
            metavar='<charge>',
            default='first',
            choices=['first', 'random', 'randomodd', 'randomeven'],
            help="""charge on the vertices.
                                    `first'  puts odd charge on first vertex;
                                    `random' puts a random charge on vertices;
                                    `randomodd' puts random odd  charge on vertices;
                                    `randomeven' puts random even charge on vertices.
                                     """)
        SimpleGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        """Build Tseitin formula according to the arguments

        Arguments:
        - `args`: command line options
        """
        G = SimpleGraphHelper.obtain_graph(args)

        if G.order() < 1:
            charge = None

        elif args.charge == 'first':

            charge = [1] + [0] * (G.order() - 1)

        else:  # random vector
            charge = [random.randint(0, 1) for _ in range(G.order() - 1)]

            parity = sum(charge) % 2

            if args.charge == 'random':
                charge.append(random.randint(0, 1))
            elif args.charge == 'randomodd':
                charge.append(1 - parity)
            elif args.charge == 'randomeven':
                charge.append(parity)
            else:
                raise ValueError(
                    'Illegal charge specification on command line')

        return TseitinFormula(G, charge)


class SCCmdHelper(FormulaHelper):
    name = 'subsetcard'
    description = 'subset cardinality formulas'

    @staticmethod
    def setup_command_line(parser):

        parser.add_argument('--equal',
                            '-e',
                            default=False,
                            action='store_true',
                            help="encode cardinality constraints as equations")

        BipartiteGraphHelper.setup_command_line(parser)

    @staticmethod
    def build_cnf(args):
        B = BipartiteGraphHelper.obtain_graph(args)
        return SubsetCardinalityFormula(B, args.equal)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of subset cardinality formulas
"""

from cnfformula.cnf import CNF

from cnfformula.graphs import bipartite_sets, enumerate_edges, neighbors


def SubsetCardinalityFormula(B, equalities=False):
    r"""SubsetCardinalityFormula

    Consider a bipartite graph :math:`B`. The CNF claims that at least half
    of the edges incident to each of the vertices on left side of :math:`B`
    must be zero, while at least half of the edges incident to each
    vertex on the left side must be one.

    Variants of these formula on specific families of bipartite graphs
    have been studied in [1]_, [2]_ and [3]_, and turned out to be
    difficult for resolution based SAT-solvers.

    Each variable of the formula is denoted as :math:`x_{i,j}` where
    :math:`\{i,j\}` is an edge of the bipartite graph. The clauses of
    the CNF encode the following constraints on the edge variables.

    For every left vertex i with neighborhood :math:`\Gamma(i)`

    .. math::
         
         \sum_{j \in \Gamma(i)} x_{i,j} \geq \frac{|\Gamma(i)|}{2}

    For every right vertex j with neighborhood :math:`\Gamma(j)`

    .. math::
         
         \sum_{i \in \Gamma(j)} x_{i,j} \leq \frac{|\Gamma(j)|}{2}.

    If the ``equalities`` flag is true, the constraints are instead
    represented by equations.
    
    .. math::
         
         \sum_{j \in \Gamma(i)} x_{i,j} = \left\lceil \frac{|\Gamma(i)|}{2} \right\rceil

    .. math::
         
         \sum_{i \in \Gamma(j)} x_{i,j} = \left\lfloor \frac{|\Gamma(j)|}{2} \right\rfloor .

    Parameters
    ----------
    B : networkx.Graph
        the graph vertices must have the 'bipartite' attribute
        set. Left vertices must have it set to 0 and the right ones to 1.
        A KeyException is raised otherwise.

    equalities : boolean
        use equations instead of inequalities to express the
        cardinality constraints.  (default: False) 

    Returns
    -------
    A CNF object

    References
    ----------
    .. [1] Mladen Miksa and Jakob Nordstrom
           Long proofs of (seemingly) simple formulas
           Theory and Applications of Satisfiability Testing--SAT 2014 (2014)
    .. [2] Ivor Spence
           sgen1: A generator of small but difficult satisfiability benchmarks
           Journal of Experimental Algorithmics (2010)
    .. [3] Allen Van Gelder and Ivor Spence
           Zero-One Designs Produce Small Hard SAT Instances
           Theory and Applications of Satisfiability Testing--SAT 2010(2010)

    """
    Left, Right = bipartite_sets(B)

    description = "Subset cardinality formula for {0}".format(B.name)
    ssc = CNF(description=description)

    def var_name(u, v):
        """Compute the variable names."""
        if u <= v:
            return 'x_{{{0},{1}}}'.format(u, v)
        else:
            return 'x_{{{0},{1}}}'.format(v, u)

    for u in Left:
        for v in neighbors(B, u):
            ssc.add_variable(var_name(u, v))

    for u in Left:
        edge_vars = [var_name(u, v) for v in neighbors(B, u)]

        if equalities:
            for cls in CNF.exactly_half_ceil(edge_vars):
                ssc.add_clause(cls, strict=True)
        else:
            for cls in CNF.loose_majority_constraint(edge_vars):
                ssc.add_clause(cls, strict=True)

    for v in Right:
        edge_vars = [var_name(u, v) for u in neighbors(B, v)]

        if equalities:
            for cls in CNF.exactly_half_floor(edge_vars):
                ssc.add_clause(cls, strict=True)
        else:
            for cls in CNF.loose_minority_constraint(edge_vars):
                ssc.add_clause(cls, strict=True)

    return ssc



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from itertools import combinations, product, permutations
from copy import copy

from cnfformula.cnf import CNF
from cnfformula.graphs import bipartite_sets, neighbors

###
### Substitions
###


class BaseSubstitution(CNF):
    """Apply a substitution to a formula
    """
    def __init__(self, cnf, new_variables=None):
        """Build a new CNF substituting variables

        Arguments:
        - `cnf`: the original cnf
        """
        assert cnf._coherent
        self._orig_cnf = cnf
        super(BaseSubstitution, self).__init__([])
        self.header = copy(cnf.header)

        # Load original variable names
        #
        # For n variables we get
        #
        # varadditional = [None, F1, F2,..., Fn]
        # substitution  = [None, F1, F2,..., Fn, -Fn, -F(n-1), ..., -F2, -F1]
        #
        variablenames = [None] + list(self._orig_cnf.variables())
        substitutions = [None] * (2 * len(variablenames) - 1)
        varadditional = [None] * (len(variablenames))

        # Transform all possible literals
        for i in range(1, len(variablenames)):
            varadditional[i] = self.transform_variable_preamble(
                variablenames[i])
            substitutions[i] = self.transform_a_literal(True, variablenames[i])
            substitutions[-i] = self.transform_a_literal(
                False, variablenames[i])

        # Collect new variable names from the CNFs:
        # clause compression needs the variable names
        if new_variables is None:
            for i in range(1, len(variablenames)):
                for clause in varadditional[i] + substitutions[
                        i] + substitutions[-i]:
                    for _, varname in clause:
                        self.add_variable(varname)
        else:
            for v in new_variables:
                self.add_variable(v)

        # Compress substitution cnfs
        for i in range(1, len(varadditional)):
            varadditional[i] = [
                list(self._compress_clause(cls)) for cls in varadditional[i]
            ]

        for i in range(1, len(substitutions)):
            substitutions[i] = [
                list(self._compress_clause(cls)) for cls in substitutions[i]
            ]

        # build and add new clauses
        for orig_cls in self._orig_cnf._clauses:

            # a substituted clause is the OR of the substituted literals
            domains = [substitutions[lit] for lit in orig_cls]
            domains = tuple(domains)

            block = [
                tuple([lit for clause in clause_tuple for lit in clause])
                for clause_tuple in product(*domains)
            ]

            self._add_compressed_clauses(block)

        # transformation may need additional clauses per variables
        for block in varadditional[1:]:
            if block:
                self._add_compressed_clauses(block)

        assert self._orig_cnf._check_coherence()
        assert self._check_coherence()

    def add_transformation_description(self, description):
        i = 1
        while 'transformation {}'.format(i) in self.header:
            i += 1
        self.header['transformation {}'.format(i)] = description

    def transform_variable_preamble(self, name):
        """Additional clauses for each variable

        Arguments:
        - `name`:     variable to add clauses for

        Returns: a list of clauses
        """
        return []

    def transform_a_literal(self, polarity, name):
        """Substitute a literal with the transformation function

        Arguments:
        - `polarity`: polarity of the literal
        - `name`:     variable to be substituted

        Returns: a list of clauses
        """
        return [[(polarity, name)]]


class IfThenElseSubstitution(BaseSubstitution):
    """Transformed formula: substitutes variable with a three variables
    if-then-else
    """
    def __init__(self, cnf):
        """Build a new CNF obtained by substituting an if-then-else to the
        variables of the original CNF

        Arguments:
        - `cnf`: the original cnf
        """
        super(IfThenElseSubstitution, self).__init__(cnf)

        self.add_transformation_description(
            "If-Then-Else substitution formula")

    def transform_a_literal(self, polarity, varname):
        """Substitute a positive literal with an if then else statement,

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: variable to be substituted

        Returns: a list of clauses
        """
        X = "{{{}}}^0".format(varname)
        Y = "{{{}}}^1".format(varname)
        Z = "{{{}}}^2".format(varname)

        return [[(False, X), (polarity, Y)], [(True, X), (polarity, Z)]]


class MajoritySubstitution(BaseSubstitution):
    """Transformed formula: substitutes variable with a Majority
    """
    def __init__(self, cnf, rank):
        """Build a new CNF obtained by substituting a Majority to the
        variables of the original CNF

        Arguments:
        - `cnf`: the original cnf
        - `rank`: how many variables in each or
        """
        self._rank = rank

        super(MajoritySubstitution, self).__init__(cnf)

        self.add_transformation_description(
            "Substitution with Majority of {}".format(self._rank))

    def transform_a_literal(self, polarity, varname):
        """Substitute a positive literal with Loose Majority,
        and negative literals with Strict Minority.

        Parameters
        ----------
        polarity : bool
            polarity of the literal
        varname  : string
            variable to be substituted

        Returns: a list of clauses
        """

        variables = ["{{{}}}^{}".format(varname, i) for i in range(self._rank)]

        threshold = (self._rank + 1) // 2  # loose majority
        if polarity:
            return list(self.greater_or_equal_constraint(variables, threshold))
        else:
            return list(self.less_than_constraint(variables, threshold))


class OrSubstitution(BaseSubstitution):
    """Transformed formula: substitutes variable with a OR
    """
    def __init__(self, cnf, rank):
        """Build a new CNF obtained by substituting a OR to the
        variables of the original CNF

        Arguments:
        - `cnf`: the original cnf
        - `rank`: how many variables in each or
        """
        self._rank = rank

        super(OrSubstitution, self).__init__(cnf)

        self.add_transformation_description(
            "Substitution with OR of {}".format(self._rank))

    def transform_a_literal(self, polarity, varname):
        """Substitute a positive literal with an OR,
        and negative literals with its negation.

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        names = ["{{{}}}^{}".format(varname, i) for i in range(self._rank)]
        if polarity:
            return [[(True, name) for name in names]]
        else:
            return [[(False, name)] for name in names]


class AllEqualSubstitution(BaseSubstitution):
    """Transformed formula: substitutes variable with 'all equals'
    """
    def __init__(self, cnf, rank):
        """Build a new CNF obtained by substituting 'all equals' to the
        variables of the original CNF

        Arguments:
        - `cnf`: the original cnf
        - `rank`: how many variables in each or
        """
        self._rank = rank

        super(AllEqualSubstitution, self).__init__(cnf)

        self.add_transformation_description(
            "Substitution with Equality of {}".format(self._rank))

    def transform_a_literal(self, polarity, varname):
        """Substitute a positive literal with an 'all equal' statement,

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        names = ["{{{}}}^{}".format(varname, i) for i in range(self._rank)]
        pairs = permutations(names, 2)
        if polarity:
            return [[(False, a), (True, b)] for a, b in pairs
                    ]  # a true variable implies all the others to true.

        else:
            return [[(False, a) for a in names],
                    [(True, a)
                     for a in names]]  # at least a true and a false variable.


class NotAllEqualSubstitution(BaseSubstitution):
    """Transformed formula: substitutes variable with 'not all equals'
    """
    def __init__(self, cnf, rank):
        """Build a new CNF obtained by substituting 'not all equals' to the
        variables of the original CNF

        Arguments:
        - `cnf`: the original cnf
        - `rank`: how many variables in each or
        """
        self._rank = rank

        super(NotAllEqualSubstitution, self).__init__(cnf)

        self.add_transformation_description(
            "Substitution with NonEquality of {}".format(self._rank))

    def transform_a_literal(self, polarity, varname):
        """Substitute a positive literal with an 'not all equal' statement,

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        return AllEqualSubstitution.transform_a_literal(
            self, not polarity, varname)


class XorSubstitution(BaseSubstitution):
    """Transformed formula: substitutes variable with a XOR
    """
    def __init__(self, cnf, rank):
        """Build a new CNF obtained by substituting a XOR to the
        variables of the original CNF

        Arguments:
        - `cnf`: the original cnf
        - `rank`: how many variables in each xor
        """
        self._rank = rank

        super(XorSubstitution, self).__init__(cnf)

        self.add_transformation_description(
            "Substitution with XOR of {}".format(self._rank))

    def transform_a_literal(self, polarity, varname):
        """Substitute a literal with a (negated) XOR

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        names = ["{{{}}}^{}".format(varname, i) for i in range(self._rank)]
        return list(self.parity_constraint(names, polarity))


class FormulaLifting(BaseSubstitution):
    """Formula lifting: Y variable select X values
    """
    def __init__(self, cnf, rank):
        """Build a new CNF obtained by lifting procedures

        Arguments:
        - `cnf`: the original cnf
        - `rank`: how many variables in each xor
        """
        self._rank = rank

        super(FormulaLifting, self).__init__(cnf)

        self.add_transformation_description(
            "Lifting with selectors over {} values".format(self._rank))

    def transform_variable_preamble(self, name):
        """Additional clauses for each lifted variable

        Arguments:
        - `name`:     variable to be substituted

        Returns: a list of clauses
        """
        selector_clauses = []
        selector_clauses.append([(True, "Y_{{{}}}^{}".format(name, i))
                                 for i in range(self._rank)])

        for s1, s2 in combinations(
            ["Y_{{{}}}^{}".format(name, i) for i in range(self._rank)], 2):
            selector_clauses.append([(False, s1), (False, s2)])

        return selector_clauses

    def transform_a_literal(self, polarity, varname):
        """Substitute a literal with a (negated) XOR

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        clauses = []
        for i in range(self._rank):
            clauses.append([(False, "Y_{{{}}}^{}".format(varname, i)),
                            (polarity, "X_{{{}}}^{}".format(varname, i))])
        return clauses


class ExactlyOneSubstitution(BaseSubstitution):
    """Transformed formula: exactly one variable is true
    """
    def __init__(self, cnf, rank):
        """Build a new CNF obtained substituting all variables with
        'exactly one' function.

        Arguments:
        - `cnf`: the original cnf
        - `rank`: how many variables in each substitution
        """
        self._rank = rank

        super(ExactlyOneSubstitution, self).__init__(cnf)

        self.add_transformation_description(
            "Substitution with Exactly one of {}".format(self._rank))

    def transform_a_literal(self, polarity, varname):
        """Substitute a literal with an \"Exactly One\"

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        clauses = []
        varnames = [
            "X_{{{}}}^{}".format(varname, i) for i in range(self._rank)
        ]

        if polarity:
            # at least one variable is true
            clauses.append([(True, name) for name in varnames])
            # no two variables are true
            for (n1, n2) in combinations(varnames, 2):
                clauses.append([(False, n1), (False, n2)])
        else:
            # if all variables but one are false, the other must be false
            for name in varnames:
                clauses.append([(False, name)] +
                               [(True, other)
                                for other in varnames if other != name])
        return clauses


class FlipPolarity(BaseSubstitution):
    """Flip the polarity of variables
    """
    def __init__(self, cnf):
        """Build a new CNF obtained by flipping the polarity of the variables

        Parameters
        ----------
        cnf : a CNF object
        """
        super(FlipPolarity, self).__init__(cnf)

        self.add_transformation_description("All polarities have been flipped")

    def transform_a_literal(self, polarity, varname):
        """Substitute a positive literal with an OR,
        and negative literals with its negation.

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        return [[(not polarity, varname)]]


class VariableCompression(BaseSubstitution):
    """Vabiable compression transformation 
    
    The original variable are substituted with the XOR (or MAJ) of
    a subset of a new set of variables (usually smaller).
    """

    _name_vertex_dict = {}
    _pattern = None
    _function = None

    def __init__(self, cnf, B, function='xor'):
        """Build a new CNF obtained by substituting a XOR to the
        variables of the original CNF.

        Parameters
        ----------
        cnf : CNF
            the original cnf formula
        B : networkx.Graph
            a bipartite graph. The right side must have the number of
            vertices equal to the number of original variables
        
        function: string 
            Select which faction is used for the compression. It must
            be one among 'xor' or 'maj'.

        """
        if function not in ['xor', 'maj']:
            raise ValueError(
                "Function specification for variable compression must be either 'xor' or 'maj'."
            )

        Left, Right = bipartite_sets(B)

        if len(Right) != len(list(cnf.variables())):
            raise ValueError(
                "Right side of the graph must match the variable numbers of the CNF."
            )

        self._pattern = B
        self._function = function
        for n, v in zip(cnf.variables(), Right):
            self._name_vertex_dict[n] = v

        super(VariableCompression, self).__init__(
            cnf, new_variables=["Y_{{{0}}}".format(i) for i in Left])

        self.add_transformation_description(
            "Variable {}-compression from {} to {} variables".format(
                function, len(Right), len(Left)))

    def transform_a_literal(self, polarity, varname):
        """Substitute a literal with a (negated) XOR

        Arguments:
        - `polarity`: polarity of the literal
        - `varname`: fariable to be substituted

        Returns: a list of clauses
        """
        varname = self._name_vertex_dict[varname]
        local_vars = neighbors(self._pattern, varname)
        local_names = ["Y_{{{0}}}".format(i) for i in local_vars]

        if self._function == 'xor':

            return list(self.parity_constraint(local_names, polarity))

        elif self._function == 'maj':

            threshold = (len(local_names) + 1) // 2  # loose majority

            if polarity:
                return list(
                    self.greater_or_equal_constraint(local_names, threshold))
            else:
                return list(self.less_than_constraint(local_names, threshold))

        else:
            raise RuntimeError(
                "Error: variable compression with invalid function")



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Implementation of formulas that check for subgraphs
"""

from cnfformula.cnf import CNF

from itertools import combinations
from itertools import product
from itertools import permutations
from cnfformula.graphs import enumerate_vertices

from math import log, ceil

from networkx import complete_graph
from networkx import empty_graph

from textwrap import dedent


def SubgraphFormula(graph, templates, symmetric=False):
    """Test whether a graph contains one of the templates.

    Given a graph :math:`G` and a sequence of template graphs
    :math:`H_1`, :math:`H_2`, ..., :math:`H_t`, the CNF formula claims
    that :math:`G` contains an isomorphic copy of at least one of the
    template graphs.

    E.g. when :math:`H_1` is the complete graph of :math:`k` vertices
    and it is the only template, the formula claims that :math:`G`
    contains a :math:`k`-clique.

    Parameters
    ----------
    graph : networkx.Graph
        a simple graph

    templates : list-like object
        a sequence of graphs.

    symmetric:
        all template graphs are symmetric wrt permutations of
        vertices. This allows some optimization in the search space of
        the assignments.

    induce: 
        force the subgraph to be induced (i.e. no additional edges are allowed)


    Returns
    -------
    a CNF object

    """
    # One of the templates is chosen to be the subgraph
    if len(templates) == 0:
        return CNF(description="Empty formula")
    elif len(templates) == 1:
        selectors = []
    elif len(templates) == 2:
        selectors = ['c']
    else:
        selectors = ['c_{{{}}}'.format(i) for i in range(len(templates))]

    # comment the formula accordingly
    if len(selectors) > 1:
        description = "{} does not contains any of the {} possible subgraphs.".format(
            graph.name, len(templates))
    else:
        description = "{} does not contain any copy of {}.".format(
            graph.name, templates[0].name)
    F = CNF(description=description)

    for s in selectors:
        F.add_variable(s)

    if len(selectors) > 1:
        for cls in F.equal_to_constraint(selectors, 1):
            F.add_clause(cls, strict=True)

    # A subgraph is chosen
    N = graph.order()
    k = max([s.order() for s in templates])

    var_name = lambda i, j: "S_{{{0},{1}}}".format(i, j)

    if symmetric:
        mapping = F.unary_mapping(list(range(k)),
                                  list(range(N)),
                                  var_name=var_name,
                                  functional=True,
                                  injective=True,
                                  nondecreasing=True)
    else:
        mapping = F.unary_mapping(list(range(k)),
                                  list(range(N)),
                                  var_name=var_name,
                                  functional=True,
                                  injective=True,
                                  nondecreasing=False)

    for v in mapping.variables():
        F.add_variable(v)

    for cls in mapping.clauses():
        F.add_clause(cls, strict=True)

    # The selectors choose a template subgraph.  A mapping must map
    # edges to edges and non-edges to non-edges for the active
    # template.

    if len(templates) == 1:

        activation_prefixes = [[]]

    elif len(templates) == 2:

        activation_prefixes = [[(True, selectors[0])], [(False, selectors[0])]]

    else:
        activation_prefixes = [[(True, v)] for v in selectors]

    # maps must preserve the structure of the template graph
    gV = enumerate_vertices(graph)

    for i in range(len(templates)):

        k = templates[i].order()
        tV = enumerate_vertices(templates[i])

        if symmetric:
            # Using non-decreasing map to represent a subset
            localmaps = product(combinations(list(range(k)), 2),
                                combinations(list(range(N)), 2))
        else:
            localmaps = product(combinations(list(range(k)), 2),
                                permutations(list(range(N)), 2))

        for (i1, i2), (j1, j2) in localmaps:

            # check if this mapping is compatible
            tedge = templates[i].has_edge(tV[i1], tV[i2])
            gedge = graph.has_edge(gV[j1], gV[j2])
            if tedge == gedge:
                continue

            # if it is not, add the corresponding
            F.add_clause(activation_prefixes[i] + \
                         [(False,var_name(i1,j1)),
                          (False,var_name(i2,j2)) ],strict=True)

    return F


def CliqueFormula(G, k):
    """Test whether a graph has a k-clique.

    Given a graph :math:`G` and a non negative value :math:`k`, the
    CNF formula claims that :math:`G` contains a :math:`k`-clique.

    Parameters
    ----------
    G : networkx.Graph
        a simple graph
    k : a non negative integer
        clique size

    Returns
    -------
    a CNF object

    """
    F = SubgraphFormula(G, [complete_graph(k)], symmetric=True)
    description = "{} does not contain any {}-clique.".format(G.name, k)
    F.header['description'] = description
    return F


def BinaryCliqueFormula(G, k):
    """Test whether a graph has a k-clique (binary encoding)

    Given a graph :math:`G` and a non negative value :math:`k`, the
    CNF formula claims that :math:`G` contains a :math:`k`-clique.
    This formula uses the binary encoding, in the sense that the
    clique elements are indexed by strings of bits.

    Parameters
    ----------
    G : networkx.Graph
        a simple graph
    k : a non negative integer
        clique size

    Returns
    -------
    a CNF object

    """
    description = "Binary {0}-clique formula".format(k)
    F = CNF(description=description)

    clauses_gen = F.binary_mapping(range(1, k + 1),
                                   G.nodes(),
                                   injective=True,
                                   nondecreasing=True)

    for v in clauses_gen.variables():
        F.add_variable(v)

    for c in clauses_gen.clauses():
        F.add_clause(c, strict=True)

    for (i1, i2), (v1, v2) in product(combinations(range(1, k + 1), 2),
                                      combinations(G.nodes(), 2)):

        if not G.has_edge(v1, v2):
            F.add_clause(clauses_gen.forbid_image(i1, v1) +
                         clauses_gen.forbid_image(i2, v2),
                         strict=True)

    return F


def RamseyWitnessFormula(G, k, s):
    """True if graph contains either k-clique or and s independent set

    Given a graph :math:`G` and a non negative values :math:`k` and
    :math:`s`, the CNF formula claims that :math:`G` contains
    a neither a :math:`k`-clique nor an independet set of size
    :math:`s`.

    Parameters
    ----------
    G : networkx.Graph
        a simple graph
    k : a non negative integer
        clique size
    s : a non negative integer
        independet set size

    Returns
    -------
    a CNF object

    """
    F = SubgraphFormula(G, [complete_graph(k), empty_graph(s)], symmetric=True)
    description = "{} does not contain {}-cliques nor {}-independent sets.".format(
        G.name, k, s)
    F.header['description'] = description



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Components for command line interface

CNFgen has many command line entry points to its functionality, and
some of them expose the same functionality over and over. This module
contains useful common components.

Copyright (C) 2012, 2013, 2014, 2015, 2016, 2019, 2020
Massimo Lauria <massimo.lauria@uniroma1.it>
https://github.com/MassimoLauria/cnfgen.git

"""

import os
import sys
import argparse
import subprocess
import tempfile
import signal

from contextlib import redirect_stdout
from contextlib import contextmanager


@contextmanager
def paginate_or_redirect_stdout(outputstream):
    """Output to a file or, when interactive, to the PAGER

    Redirect standard output to ``outputstream``. Furthermore when the
    standard output is supposed to go to an interactive terminal (i.e.
    it has not been piped to a file or to another process) then
    instead of flashing it on the screen this context manager
    redirects it to a temporary file which is shown by `$PAGER`, or by
    `less` if PAGER environment variable is not defined.
    """

    with redirect_stdout(outputstream):
        use_pager = sys.stdout.isatty()
        pager = os.getenv('PAGER', 'less')

        if use_pager:
            path = tempfile.mkstemp()[1]
            tmp_file = open(path, 'a')
        else:
            tmp_file = sys.stdout

        with redirect_stdout(tmp_file):
            yield

        if use_pager:
            tmp_file.flush()
            tmp_file.close()
            p = subprocess.Popen([pager, path], stdin=subprocess.PIPE)
            p.communicate()


@contextmanager
def redirect_stdin(stream):
    """Redirect stdin during a test
    TODO: move to contextlib.redirect_stdin once it is implemented
    """
    old_stdin = sys.stdin
    sys.stdin = stream
    yield
    sys.stdin = old_stdin


def setup_SIGINT():
    """Register a handler for SIGINT signal

    Register a handler that manages keyboard interruptions
    via SIGINT.
    """
    def sigint_handler(insignal, frame):

        progname = os.path.basename(sys.argv[0])
        signame = signal.Signals(insignal).name
        print('{} received: program \'{}\' stops.'.format(signame, progname),
              file=sys.stderr)
        sys.exit(-1)

    signal.signal(signal.SIGINT, sigint_handler)


def find_in_package(package, test, sortkey=None):
    """Explore a package for items that satisfy a speficic test"""
    import pkgutil

    result = []

    if sortkey is None:
        sortkey = str

    for loader, module_name, _ in pkgutil.walk_packages(package.__path__):
        module_name = package.__name__ + "." + module_name
        if module_name in sys.modules:
            module = sys.modules[module_name]
        else:
            module = loader.find_module(module_name).load_module(module_name)
        for objname in dir(module):
            obj = getattr(module, objname)
            if test(obj):
                result.append(obj)
    result.sort(key=sortkey)
    return result


def get_transformation_helpers():

    import cnfgen
    from .transformation_helpers import TransformationHelper

    def test(x):
        return isinstance(x, type) and \
            issubclass(x, TransformationHelper) and \
            x != TransformationHelper

    return find_in_package(cnfgen, test, sortkey=lambda x: x.name)


def get_formula_helpers():

    import cnfgen
    from .formula_helpers import FormulaHelper

    def test(x):
        return isinstance(x, type) and \
            issubclass(x, FormulaHelper) and \
            x != FormulaHelper

    return find_in_package(cnfgen, test, sortkey=lambda x: x.name)


class CLIError(Exception):
    """Error related to the command line arguments

This error occurs when the command line contains some errors. """
    pass


class CLIParser(argparse.ArgumentParser):
    """Argument Parser for CNFGen

Exactly as argparse.ArgumentParser, but the error function raises an
exception, instead of calling exit.
"""
    def __init(self,
               prog=None,
               usage=None,
               description=None,
               epilog=None,
               parents=[],
               prefix_chars='-',
               fromfile_prefix_chars=None,
               argument_default=None,
               formatter_class=argparse.RawDescriptionHelpFormatter,
               conflict_handler='error',
               add_help=True,
               allow_abbrev=True):
        super(CLIParser,
              self).__init__(prog=prog,
                             usage=usage,
                             description=description,
                             epilog=epilog,
                             parents=parents,
                             formatter_class=formatter_class,
                             prefix_chars=prefix_chars,
                             fromfile_prefix_chars=fromfile_prefix_chars,
                             argument_default=argument_default,
                             conflict_handler=conflict_handler,
                             add_help=add_help,
                             allow_abbrev=allow_abbrev)

    def error(self, message):
        message = str(message)
        errstr = ["ERROR: " + x for x in message.splitlines()]
        errstr.append("")

        if self.usage is not None:
            errstr.append(self.usage)
            errstr.append("")

        if self.description is not None:
            errstr.append(self.description)
            errstr.append("")

        errstr.append("See '{0} -h' or '{0} --help' for more info.".format(
            self.prog))

        raise CLIError("\n".join(errstr))


def positive_int(value):
    errmsg = "{} was supposed to be a positive integer".format(value)
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(errmsg)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(errmsg)
    return ivalue


def nonnegative_int(value):
    errmsg = "{} was supposed to be a non negative integer".format(value)
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(errmsg)
    if ivalue < 0:
        raise argparse.ArgumentTypeError(errmsg)
    return ivalue

def positive_even_int(value):
    errmsg = "{} was supposed to be an even integer".format(value)
    ivalue = positive_int(value)
    if ivalue % 2 != 0:
        raise argparse.ArgumentTypeError(errmsg)
    return ivalue

def probability(value):
    errmsg = "{} was supposed to be a real number in [0,1]".format(value)
    try:
        p = float(value)
    except ValueError:
        raise argparse.ArgumentTypeError(errmsg)
    if not (0 <= p <= 1.0):
        raise argparse.ArgumentTypeError(errmsg)
    return p



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""Utilities to build dimacs encoding of pebbling formulas

Accepts only the kthlist graph format:

ASSUMPTIONS: the graph is given with a line for each vertex, from
sources to a *single sink*.

CNF formulas interesting for proof complexity.
"""

import os
import sys
import argparse

from cnfformula import PebblingFormula
from cnfformula import readGraph

from .cmdline import paginate_or_redirect_stdout
from .cmdline import redirect_stdin
from .cmdline import setup_SIGINT
from .cmdline import get_transformation_helpers
from .cmdline import CLIParser, CLIError

from .msg import interactive_msg
from .msg import error_msg
from .msg import msg_prefix
from .msg import InternalBug

#################################################################
#          Command line tool follows
#################################################################


# Command line helpers
def setup_command_line(parser):
    """Setup general command line options

    Arguments:
    - `parser`: parser to fill with options
    """
    parser.add_argument('--output',
                        '-o',
                        type=argparse.FileType('w'),
                        metavar="<output>",
                        default='-',
                        help="""Output file. The formula is saved
                        on file instead of being sent to standard
                        output. Setting '<output>' to '-' is another
                        way to send the formula to standard output.
                        (default: -)
                        """)
    parser.add_argument(
        '--input',
        '-i',
        type=argparse.FileType('r'),
        metavar="<input>",
        default='-',
        help=
        """Input file. The DAG is read from a file instead of being read from
                        standard output. Setting '<input>' to '-' is
                        another way to read from standard
                        input.  (default: -)
                        """)
    parser.add_argument('--quiet',
                        '-q',
                        action='store_false',
                        dest='verbose',
                        help="""Output just the formula with no header.""")

    subparsers = parser.add_subparsers(title="Available transformation",
                                       metavar="<transformation>")

    for sc in get_transformation_helpers():
        p = subparsers.add_parser(sc.name, help=sc.description)
        sc.setup_command_line(p)
        p.set_defaults(transformation=sc)


# Main program
def cli(argv=sys.argv, mode='output'):
    """From KTHLists to Pebbling formulas

    Parameters
    ----------
    argv: list, optional
        The list of token with the command line arguments/options.

    mode: str
        One among 'formula', 'string', 'output' (latter is the default)
        - 'formula' return a CNF object
        - 'string' return the string with the output of CNFgen
        - 'output' output the formula to the user
    """

    # Parse the command line arguments
    parser = CLIParser(prog=os.path.basename(argv[0]))
    setup_command_line(parser)

    # Be lenient on non string arguments
    argv = [str(x) for x in argv]

    args = parser.parse_args(argv[1:])

    ask_kthlist_graph = """
       Waiting for a directed acyclic graph on <stdin>,
       in 'kthlist' format.

       See: https://massimolauria.net/cnfgen/graphformats.html"""

    with redirect_stdin(args.input), msg_prefix('c '):

        with msg_prefix('GRAPH INPUT: '):
            interactive_msg(ask_kthlist_graph)

        G = readGraph(sys.stdin, "dag", file_format="kthlist")

    F = PebblingFormula(G)

    if hasattr(args, "transformation"):
        F2 = args.transformation.transform_cnf(F, args)
    else:
        F2 = F

    if mode == 'formula':
        return F2
    elif mode == 'string':
        return F2.dimacs(export_header=args.verbose)
    else:
        with paginate_or_redirect_stdout(args.output):
            print(F2.dimacs(export_header=args.verbose))


# Launcher
def main():
    setup_SIGINT()
    try:

        command_line_utility(sys.argv)

    except ValueError as e:
        error_msg("GRAPH ERROR: " + str(e))
        sys.exit(-1)

    except CLIError as e:
        error_msg(str(e))
        sys.exit(-1)

    except InternalBug as e:
        print(str(e), file=sys.stderr)
        sys.exit(-1)


if __name__ == '__main__':
    main()
